# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Logging and Event utils"""

import logging
import sys
import os
import json
import tempfile
import watchtower
import numpy as np
from pathlib import Path


class Logger:
    """
    Logger utility. Provides a thin wrapper over the builtin
    python logging utility.

    Attributes:
        log_level (str): logging level.
        service (str): service name.
        _terminal_logger: logger instance that prints to console.
    """

    def __init__(
        self, log_level="INFO", log_file=None, log_cloudwatch=False, service="Testing"
    ):
        self.service = service
        self._terminal_logger = setup_logger(
            log_level, log_file=log_file, log_cloudwatch=log_cloudwatch, service=service
        )

    def info(self, *args):
        self._terminal_logger.info(args)

    def error(self, *args):
        self._terminal_logger.error(args)

    def warning(self, *args):
        self._terminal_logger.warning(args)

    def critical(self, *args):
        self._terminal_logger.critical(args)


class WorkflowOutputs:
    """
    Supports writing outputs to local file system
    """

    def get_local_file_size(self, filename):
        """
        Basic function get the local file size. Returns -1 for non-existent files

        Parameters
        ----------
        filename: str
            Full name of the file (path).
        """
        try:
            file_size = os.path.getsize(filename)
        except FileNotFoundError:
            file_size = -1
        return file_size

    def split_workflow_outputs(self, path_str, result):
        """
        Function to split workflow outputs into individual files on local file system.

        Args:
            path (str): outputs base file path
            result (dict): single processing result data
        """

        path = Path(path_str)

        filename_path = path / "filename.txt"
        size_path = path / "size.txt"
        run_id_path = path / "run_id.txt"
        action_path = path / "action.txt"
        status_path = path / "status.txt"

        filename_path.write_text(result["filename"])
        size_path.write_text(str(result["size"]))
        run_id_path.write_text(str(result["run_id"]))
        action_path.write_text(result["action"])
        status_path.write_text(result["status"])

    def put_generate_outputs_list(self, output_data, generate_status):
        """
        Function create generate outputs list json file on local file system.

        Args:
            output_data (dict): data payload for output.
            generate_status (str): starting|complete|error.
        """

        tmp_dir = Path(tempfile.gettempdir())
        output_file_path = tmp_dir / "results.json"
        results = []

        file_list = output_data["file_list"]

        for index, filename in enumerate(file_list):
            size = self.get_local_file_size(filename)
            if size < 0:
                raise FileNotFoundError()

            result = {
                "filename": filename,
                "size": size,
                "result": output_data["result"],
                "status": generate_status,
            }

            results.append(result)

        output_file_path.write_text(json.dumps(results))

    def put_workflow_outputs(self, output_data, action):
        """
        Function to create workflow outputs json file on local file system.

        Args:
            output_data (dict): single processing result data
            action (str): workflow action
        """

        tmp_dir = Path(tempfile.gettempdir())
        output_file_path = tmp_dir / "results.json"

        # Most operating systems should have a temp directory
        if not tmp_dir.exists:
            try:
                tmp_dir.mkdir()
            except OSError:
                print("creation of temp directory failed")

        size = self.get_local_file_size(output_data["filename"])
        if size < 0:
            raise FileNotFoundError()

        result = {
            "filename": output_data["filename"],
            "size": size,
            "run_id": output_data["run_id"],
            "action": action,
            "status": output_data["result"],
        }

        # Argo limitations require outputting each value separately
        self.split_workflow_outputs(str(tmp_dir), result)

        output_file_path.write_text(json.dumps(result))

    def put_workflow_outputs_list(self, output_data, action):
        """
        Function to create workflow outputs list json file on local file system.

        Args:
            output_data (list[dict]): processing result data
            action (str): workflow action
        """

        tmp_dir = Path(tempfile.gettempdir())
        output_file_path = tmp_dir / "results.json"
        results = []

        # Most operating systems should have a temp directory
        if not tmp_dir.exists:
            try:
                tmp_dir.mkdir()
            except OSError:
                print("creation of temp directory failed")

        file_list = output_data["file_list"]

        for index, filename in enumerate(file_list):
            size = self.get_local_file_size(filename)
            if size < 0:
                raise FileNotFoundError()

            result = {
                "filename": filename,
                "size": size,
                "run_id": output_data["run_list"][index],
                "action": action,
                "status": output_data["result_list"][index],
            }

            results.append(result)

        output_file_path.write_text(json.dumps(results))


def setup_logger(
    log_level="INFO",
    log_file=None,
    log_cloudwatch=False,
    service="Testing",
    np_precision=3,
):
    """
    Creates and configures a logger object.

    Args:
        log_level (str): the logging level. Defaults to 'INFO'.
        log_file (str): optional local log_file. Defaults to None.
        log_cloudwatch (bool): Will log to CloudWatch topic if set to True.
            Defaults to False.
        service (str): service name .
        np_precision (int): numpy precision.

    Returns:
        logging.Logger: python Logger object.
    """

    # numpy float precisions when printing
    np.set_printoptions(precision=np_precision)

    # Python logger config
    logger = logging.getLogger(
        __name__ + "." + service
    )  # '' or None is the root logger

    # Remove all previous filters and handlers
    logger.handlers = []
    logger.filters = []

    # Get handler on log
    if log_file is not None:
        hdlr = logging.FileHandler(log_file, "a")
    else:
        hdlr = logging.StreamHandler(sys.stdout)

    logger.addHandler(hdlr)

    # Add cloudwatch logging if requested
    if log_cloudwatch:
        hdlr = watchtower.CloudWatchLogHandler()

    logger.addHandler(hdlr)

    fmt_str = "# {cols[y]}%(asctime)s{cols[reset]}"
    fmt_str += " %(levelname)-8s"
    fmt_str += " {cols[c]}%(funcName)-3s{cols[reset]} %(message)s"

    logger.setLevel(1 if log_level == "ALL" else getattr(logging, log_level))
    logger.propagate = False
    return logger



#!/usr/bin/env python3
#  Copyright (c) 2019 Toyota Research Institute

"""Script for joining together two Maccor files that correspond to the same run
but data has been split between two files. The column increment function makes sure
that certain columns are monotonically increasing across the two files. The metadata
line from the first file is used in the final output file

Usage:
    splice.py [options]
    splice.py (-h | --help)

Options:
    -h --help                Show this screen
    --version                Show version

"""
import pandas as pd
from beep import StringIO


class MaccorSplice:
    def __init__(self, input_1, input_2, output):
        """
        Args:
            input_1 (str): Filename corresponding to first file.
            input_2 (str): Filename corresponding to second file.
            output (str): Filename of output file.
        """
        self.input_1 = input_1
        self.input_2 = input_2
        self.output = output

    def read_maccor_file(self, filename):
        """
        Loads Maccor file and returns meta data line and data frame containing
        file data.

        Args:
            filename (str): path to file.

        Returns:
            str: first line of maccor file containing meta data.
            pandas.DataFrame: data frame with file data.

        """
        with open(filename) as f:
            lines = f.readlines()
        metadata_line = lines[0]
        data_lines = lines[1:]

        # Parse data
        data_text = "\n".join(data_lines)
        tsv_part = pd.read_csv(StringIO(data_text), delimiter="\t")
        return metadata_line, tsv_part

    def write_maccor_file(self, metadata_line, dataframe, output):
        """
        Writes data and meta data into a Maccor file.

        Args:
            metadata_line (str): line containing meta data.
            dataframe (pandas.DataFrame): content data.
            output (str): output file name.
            output (str): output file name.
        """
        with open(output, "w") as write_tsv:
            write_tsv.writelines(metadata_line)
            write_tsv.write(dataframe.to_csv(sep="\t", index=False))

    def column_increment(self, data_1, data_2):
        """
        Special increment logic.

        Args:
            data_1 (pandas.DataFrame):
            data_2 (pandas.DataFrame):

        Returns:
            pandas.DataFrame: data_1 transformed (incremented)
            pandas.DataFrame: data_2 transformed (incremented)
        """
        columns_to_update = [
            "Rec#",
            "Cyc#",
            "Test (Sec)",
            "Loop1",
            "Loop2",
            "Loop3",
            "Loop4",
        ]
        for column in columns_to_update:
            if data_2[column].iloc[0] < data_1[column].iloc[-1]:
                data_2[column] = data_2[column] + data_1[column].iloc[-1]

        return data_1, data_2

    def splice_operation(self, data_1, data_2):
        """
        Concatenates two data frames.

        Args:
            data_1 (pandas.DataFrame):
            data_2 (pandas.DataFrame):

        Returns:
            pandas.DataFrame: concatenated data frame.

        """
        data_final = pd.concat([data_1, data_2])
        return data_final

    def run_splice(self):
        """
        Reads two input maccor files. Concatenates the respective data frames.
        Writes to a new Maccor file.
        """

        metadata_line_1, data_1 = self.read_maccor_file(self.input_1)
        metadata_line_2, data_2 = self.read_maccor_file(self.input_2)
        data_1, data_2 = self.column_increment(data_1, data_2)
        data_final = self.splice_operation(data_1, data_2)
        self.write_maccor_file(metadata_line_1, data_final, self.output)



"""Classes and functions for handling Battery Archive cycler data.

"""
import os
from datetime import datetime

import pytz
import pandas as pd
import numpy as np

from beep.structure.base import BEEPDatapath


class BatteryArchiveDatapath(BEEPDatapath):
    """A datapath for Battery Archive cycler data.


    One file is required - a *timeseries.csv data file
    from the Battery Archive. No metadata is supported.

    Attributes:
        All from BEEPDatapath
    """

    # Mapping of raw data file columns to BEEP columns
    COLUMN_MAPPING = {
        "test_time (s)": "test_time",
        "cycle_index": "cycle_index",
        "current (a)": "current",
        "voltage (v)": "voltage",
        "charge_capacity (ah)": "charge_capacity",
        "discharge_capacity (ah)": "discharge_capacity",
        "charge_energy (wh)": "charge_energy",
        "discharge_energy (wh)": "discharge_energy",
        "cell_temperature (c)": "temperature",
        "date_time": "date_time"
    }

    # Columns to ignore
    COLUMNS_IGNORE = ["environment_temperature (c)"]

    # Mapping of data types for BEEP columns
    DATA_TYPES = {
        "test_time": "float64",
        "cycle_index": "int32",
        "current": "float32",
        "voltage": "float32",
        "charge_capacity": "float64",
        "discharge_capacity": "float64",
        "charge_energy": "float64",
        "discharge_energy": "float64",
        "temperature": "float32",
        "date_time": "float32",
    }

    FILE_PATTERN = ".*timeseries\\.csv"

    @classmethod
    def from_file(cls, path):
        """Load a Battery Archive cycler file from raw file.

        Step indices and times are not given, so they are reverse engineered
        based on current. Three main steps are assigned for each cycle:

        1. Rest
        2. Charge
        3. Discharge

        Args:
            path (str, Pathlike): Path to the raw data csv.

        Returns:
            (ArbinDatapath)
        """
        df = pd.read_csv(path)
        df.rename(str.lower, axis="columns", inplace=True)
        df.drop(columns=[c for c in cls.COLUMNS_IGNORE if c in df.columns], inplace=True)
        df["step_index"] = 0

        df["step_index"] = df["current (a)"].apply(decide_step_index)

        step_change_ix = np.where(np.diff(df["step_index"], prepend=np.nan))[0]

        # get list of start times to subtract
        start_times = df["test_time (s)"].loc[step_change_ix]
        arrays = [None] * len(step_change_ix)

        final_ix = [df.shape[0] - 1]
        step_change_ix_buffered = np.append(step_change_ix[1:], final_ix)

        for i, scix in enumerate(step_change_ix_buffered):
            prev_scix = step_change_ix[i]
            n_repeats = scix - prev_scix
            arrays[i] = np.repeat(start_times.loc[prev_scix], n_repeats)

        # include the final data point, which is not a real step change
        subtractor = np.concatenate(arrays, axis=0)
        subtractor = np.append(subtractor, [subtractor[-1]])

        df["step_time"] = df["test_time (s)"] - subtractor

        df.rename(columns=cls.COLUMN_MAPPING, inplace=True)
        dtfmt = '%Y-%m-%d %H:%M:%S.%f'
        # convert date time string to
        dts = df["date_time"].apply(lambda x: datetime.strptime(x, dtfmt))

        df["date_time"] = dts.apply(lambda x: x.timestamp())
        df["date_time_iso"] = dts.apply(lambda x: x.replace(tzinfo=pytz.UTC).isoformat())

        for column, dtype in cls.DATA_TYPES.items():
            if column in df:
                if not df[column].isnull().values.any():
                    df[column] = df[column].astype(dtype)

        paths = {
            "raw": os.path.abspath(path),
            "metadata": None
        }

        # there is no metadata given in the BA files
        metadata = {}

        return cls(df, metadata, paths)


def decide_step_index(i):
    """
    Decide a step index based on current values.
    Args:
        i (float): Current value

    Returns:
        (int): Step index
    """
    if np.abs(i) < 1e-6:
        return 1
    elif i < 0:
        return 3
    else:
        return 2



"""Classes and functions for handling BioLogic battery cyclers.
"""

import hashlib
from datetime import datetime

import pandas as pd

from beep.structure.base import BEEPDatapath
from beep.conversion_schemas import BIOLOGIC_CONFIG


class BiologicDatapath(BEEPDatapath):
    """Datapath for ingesting and structuring BioLogic cycler data.
    """

    @classmethod
    def from_file(cls, path):
        """Creates a BEEPDatapath from a raw BioLogic battery cycler output file.

        Args:
            path (str): file path to data file

        Returns:
            BiologicDatapath
        """

        header_line = 1  # specific to file layout
        data_starts_line = 2  # specific to file layout
        column_map = BIOLOGIC_CONFIG["data_columns"]

        raw = dict()
        i = 0
        with open(path, "rb") as f:

            empty_lines = 0  # used to find the end of the data entries.
            while empty_lines < 2:  # finding 2 empty lines in a row => EOF
                line = f.readline()
                i += 1

                if i == header_line:
                    header = str(line.strip())[2:-1]
                    columns = header.split(";")
                    for c in columns:
                        raw[c] = list()

                if i >= data_starts_line:
                    line = line.strip().decode()
                    if len(line) == 0:
                        empty_lines += 1
                        continue
                    items = line.split(";")
                    for ci in range(len(items)):
                        column_name = columns[ci]
                        data_type = column_map.get(column_name, dict()).get(
                            "data_type", str
                        )
                        scale = column_map.get(column_name, dict()).get("scale", 1.0)
                        item = items[ci]
                        if data_type == "int":
                            item = int(float(item))
                        if data_type == "float":
                            item = float(item) * scale
                        raw[column_name].append(item)

        data = dict()
        for column_name in column_map.keys():
            data[column_map[column_name]["beep_name"]] = raw[column_name]
        data["data_point"] = list(range(1, len(raw["cycle number"]) + 1))

        data = pd.DataFrame(data)
        data.loc[data.step_index % 2 == 0, "charge_capacity"] = abs(
            data.charge_capacity
        )
        data.loc[data.step_index % 2 == 1, "charge_capacity"] = 0
        data.loc[data.step_index % 2 == 1, "discharge_capacity"] = abs(
            data.discharge_capacity
        )
        data.loc[data.step_index % 2 == 0, "discharge_capacity"] = 0

        metadata_path = path.replace(".csv", ".mpl")
        metadata = cls.parse_metadata(metadata_path)
        metadata["filename"] = path
        metadata["_today_datetime"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        paths = {
            "raw": path,
            "metadata": metadata_path
        }

        return cls(data, metadata, paths)

    @staticmethod
    def parse_metadata(metadata_path):
        """Extracts BioLogic metadata from metadata file.

        Iterates through a biologic metadata file and extracts the meta fields:
            - cell_id
            - barcode
            - protocol

        Args:
            metadata_path (str): path to metadata file

        Returns:
            (dict): dictionary with metadata fields
        """

        flag_cell_id = False
        flag_barcode = False
        flag_protocol = False
        flag_protocol_start = False
        protocol_text = ""
        max_lines = 10000

        metadata = dict()

        with open(metadata_path, "rb") as f:

            i = 0
            while True:
                line = f.readline()
                line = str(line)
                if line.startswith("b'"):
                    line = line[2:]
                if line.endswith("'"):
                    line = line[:-1]
                if line.endswith("\\r\\n"):
                    line = line[:-4]

                if line.strip().split(" : ")[0] == "Run on channel":
                    channel_id = line.strip().split(' : ')[1]
                    channel_id = channel_id.split(' ')[0]
                    channel_id = int(str(ord(channel_id[0]) - 64) + channel_id[1:])
                    metadata["channel_id"] = channel_id
                    flag_cell_id = True

                if "Comments" in line.strip().split(" : ")[0]:
                    barcode = line.strip().split(" : ")[-1]
                    metadata["barcode"] = barcode
                    flag_barcode = True

                if "Cycle Definition" in line.strip().split(" : ")[0]:
                    protocol_text += line + "\n"
                    flag_protocol_start = True

                if flag_protocol_start:
                    if line == "":
                        flag_protocol = True
                        protocol = hashlib.md5(protocol_text.encode()).digest()
                        metadata["protocol"] = protocol
                    else:
                        protocol_text += line + "\n"

                done = flag_cell_id and flag_barcode and flag_protocol
                if done:
                    break

                i += 1
                if i > max_lines:
                    break
        return metadata



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and script for renaming cycler files.

Usage:
    collate

Options:
    -h --help       Show this screen
    --version       Show version

The `collate` script takes no input, and operates by assuming the BEEP_PROCESSING_DIR (default `/`)
has subdirectories `/data-share/raw_cycler_files` and `data-share/renamed_cycler_files/FastCharge`.

The script moves files from the `/data-share/raw_cycler_files` directory, parses the metadata,
and renames them according to a combination of protocol, channel number, and date, placing them in
`/data-share/renamed_cycler_files`.

The script output is a json string that contains the following fields:

* `fid` - The file id used internally for renaming
* `filename` - full paths for raw cycler filenames
* `strname` - the string name associated with the file (i. e. scrubbed of `csv`)
* `file_list` - full paths for the new, renamed, cycler files
* `protocol` - the cycling protocol corresponding to each file
* `channel_no` - the channel number corresponding to each file
* `date` - the date corresponding to each file

Example:
>>> collate
... {"fid": [0, 1, 2],
...  "strname": ["2017-05-09_test-TC-contact",
...              "2017-08-14_8C-5per_3_47C",
...              "2017-12-04_4_65C-69per_6C"],
...  "file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
...                "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv",
...                "/data-share/renamed_cycler_files/FastCharge/FastCharge_2_CH29.csv"],
...  "protocol": [null, "8C(5%)-3.47C", "4.65C(69%)-6C"], "date": ["2017-05-09", "2017-08-14", "2017-12-04"],
...  "channel_no": ["CH33", "CH44", "CH29"],
...  "filename": ["/data-share/raw_cycler_files/2017-05-09_test-TC-contact_CH33.csv",
...               "/data-share/raw_cycler_files/2017-08-14_8C-5per_3_47C_CH44.csv",
...               "/data-share/raw_cycler_files/2017-12-04_4_65C-69per_6C_CH29.csv"]}
"""


import os
import re
import shutil
import json
import warnings

import pandas as pd
from monty.serialization import dumpfn
from beep import tqdm


SRC_DIR = os.path.join("data-share", "raw_cycler_files")
DEST_DIR = os.path.join("data-share", "renamed_cycler_files")
PROJECT_NAME = "FastCharge"
METADATA_COLUMN_NAMES = [
    "fid",
    "protocol",
    "channel_no",
    "date",
    "strname",
    "filename",
    "file_list",
]


def get_parameters_fastcharge(filename, source_directory):
    """
    Parse the filename to get parameters out, with fallback to metadata file

    Args:
        filename (str): this is the name of the file to be renamed.
        source_directory (str): this is the path to the file to be renamed

    Returns:
        Returns the date, the channel number, the name of the file, and the protocol (CC1, Q1, CC2) as str.

    """
    filename = filename.rsplit(".", 1)[0]
    strname = filename.rsplit("_CH", 1)[0]

    # Get date
    date = re.match(r"(\d+)-(\d+)-(\d+)", strname)
    if date is None:
        warnings.warn("Date could not be parsed from {}".format(filename))
    else:
        date = date.group()

    # Get channel number
    chno = re.findall(r"(CH+\d*)", filename)
    if chno:
        chno = chno[0]
    else:
        chno = None
        warnings.warn("Channel number could not be parsed from {}".format(filename))

    try:
        param_str = strname.rsplit(date + "_", 1)[1]
        param = param_str.replace("_", ".")

        if param.find(".") < 0:
            param = find_meta(filename, source_directory)

        q1 = re.findall(r"(\d+per)", param)[0].split("per")[0]
        cc1 = re.findall(r"(\d*\.?\d+C)", param.split(q1 + "per.")[0])[0]
        cc2 = re.findall(r"(\d*\.?\d+C)", param.split(q1 + "per.")[1])[0]
        protocol = cc1 + "(" + q1 + "%)-" + cc2
    except Exception as e:
        warnings.warn("Failed to parse protocol for {}: {}".format(filename, e))
        protocol = None

    return date, chno, strname, protocol


def get_parameters_oed(filename, source_directory):
    """
    Parse the filename to get parameters out, with fallback to the metadata file

    Args:
        filename (str): this is the name of the file to be renamed.
        source_directory (str): this is the path to the file to be renamed.

    Returns:
        str: Returns the date, the channel number, the name of the file, and the protocol (CC1, Q1, CC2) as str.

    """
    filename = filename.rsplit(".", 1)[0]
    strname = filename.rsplit("_CH", 1)[0]

    # Get date
    date = re.match(r"(\d+)-(\d+)-(\d+)", strname)
    if date is None:
        warnings.warn("Date could not be parsed from {}".format(filename))
    else:
        date = date.group()

    # Get channel number
    chno = re.findall(r"(CH+\d*)", filename)
    if chno:
        chno = chno[0]
    else:
        chno = None
        warnings.warn("Channel number could not be parsed from {}".format(filename))

    try:
        param_str = strname.rsplit(date + "_", 1)[1]
        param = param_str.replace("_", ".")

        if param.find(".") < 0 or param.find("oed") >= 0:
            param = find_meta(filename, source_directory)

        cc1 = param.split(".")[0].lower().replace("pt", ".")
        cc2 = param.split(".")[1].lower().replace("pt", ".")
        cc3 = param.split(".")[2].lower().replace("pt", ".")
        cc4 = param.split(".")[3].lower().replace("pt", ".")
        protocol = {"cc1": cc1, "cc2": cc2, "cc3": cc3, "cc4": cc4}
        protocol = json.dumps(protocol)
    except Exception as e:
        warnings.warn("Failed to parse protocol for {}: {}".format(filename, e))
        protocol = None

    return date, chno, strname, protocol


def find_meta(filename, source_directory):
    """
    Find the metadata associated with the filename, and return the parameter string.
    Useful for file names such as batch#.

    Args:
        filename (str): this is the name of the file to be renamed.
        source_directory (str): this is the path to the file to be renamed.

    Returns:
        str: string containing parameters, to be parsed in GetParameters.

    """
    metafile = os.path.join(source_directory, filename + "_Metadata.csv")
    metadf = pd.read_csv(metafile)
    metadf = metadf.rename(str.lower, axis="columns")

    schfile = (
        metadf["schedule_file_name"][0].split("\\")[-1].split(".sdu")[0].split("-")[1]
    )
    param = schfile.replace("_", ".")

    return param


def init_map(project_name, destination_directory):
    """
    Create the initial mapping file if the project is new; import the existing mapping file
    if the project has already been renamed.

    Args:
        project_name (str): name of the project.
        destination_directory (str): directory to save the renamed files to.

    Returns:
        str: unique file identifier (file_id).
        pandas.DataFrame : the dataframe of the mapping file.

    """
    project_path = os.path.join(destination_directory, project_name)
    map_filename = os.path.join(project_path, project_name + "map.csv")
    if not os.path.exists(project_path):
        os.makedirs(os.path.join(destination_directory, project_name))
        file_id = 0
        mapdf = pd.DataFrame(columns=METADATA_COLUMN_NAMES)
        open(map_filename, "a").close()
    elif len(os.listdir(project_path)) == 1:
        file_id = 0
        mapdf = pd.DataFrame(columns=METADATA_COLUMN_NAMES)
    else:
        mapdf = pd.read_csv(map_filename)
        mapdf.columns = METADATA_COLUMN_NAMES
        file_id = mapdf["fid"].max() + 1
    return file_id, mapdf


def process_files_json():
    """
    Inspects the BEEP_PROCESSING_DIR directory and renames
    files according to the prescribed system of protocol/date/run ID
    associated with the file metadata.  Since this script operates
    only on filesystem assumptions, no input is required.

    Returns:
        (str): json string corresponding to the locations of the renamed files.
    """
    # chdir into beep root
    pwd = os.getcwd()
    os.chdir(os.environ.get("BEEP_PROCESSING_DIR", "/"))

    if not os.path.exists(SRC_DIR):
        os.makedirs(SRC_DIR)
    if not os.path.exists(DEST_DIR):
        os.makedirs(DEST_DIR)

    meta_list = list(filter(lambda x: "_Metadata.csv" in x, os.listdir(SRC_DIR)))
    file_list = list(
        filter(
            lambda x: ".csv" in x if x not in meta_list else None, os.listdir(SRC_DIR)
        )
    )
    all_list = list(filter(lambda x: ".csv" in x, os.listdir(SRC_DIR)))

    all_list = sorted(all_list)
    dumpfn(all_list, "all_files.json")

    [file_id, mapdf] = init_map(PROJECT_NAME, DEST_DIR)

    new_file_index = file_id

    for filename in tqdm(sorted(file_list)):
        # If the file has already been renamed another entry should not be made
        if mapdf["filename"].str.contains(filename).sum() > 0:
            continue
        old_file = os.path.join(SRC_DIR, filename)
        new_path = os.path.join(DEST_DIR, PROJECT_NAME)
        shutil.copy(old_file, new_path)  # copy main data file
        shutil.copy(
            old_file.replace(".csv", "_Metadata.csv"), new_path
        )  # copy meta data file

        if PROJECT_NAME == "FastCharge":
            [date, channel_no, strname, protocol] = get_parameters_fastcharge(
                filename, SRC_DIR
            )
        elif PROJECT_NAME == "ClosedLoopOED":
            [date, channel_no, strname, protocol] = get_parameters_oed(
                filename, SRC_DIR
            )
        else:
            raise ValueError("Unsupported PROJECT_NAME: {}".format(PROJECT_NAME))

        # Look for an existing row with the protocol and date and use that row if it
        # already exists
        row = mapdf[(mapdf["protocol"] == protocol) & (mapdf["date"] == date)]
        if len(row) > 0:
            file_id = row["fid"].iloc[0]
            protocol = row["protocol"].iloc[0]
            date = row["date"].iloc[0]
            strname = row["strname"].iloc[0]
        # Otherwise create a new file
        else:
            file_id = new_file_index
            new_file_index = new_file_index + 1
        new_name = "{}_{}_{}".format(PROJECT_NAME, f"{file_id:06}", channel_no)
        new_file = os.path.join(DEST_DIR, PROJECT_NAME, "{}.csv".format(new_name))

        new_row = pd.DataFrame(
            [
                [
                    file_id,
                    protocol,
                    channel_no,
                    date,
                    strname,
                    os.path.abspath(old_file),
                    os.path.abspath(new_file),
                ]
            ],
            columns=METADATA_COLUMN_NAMES,
        )
        mapdf = mapdf.append(new_row)

        os.rename(os.path.join(DEST_DIR, PROJECT_NAME, filename), new_file)
        os.rename(
            os.path.join(DEST_DIR, PROJECT_NAME, filename).replace(
                ".csv", "_Metadata.csv"
            ),
            new_file.replace(".csv", "_Metadata.csv"),
        )

        mapdf.to_csv(
            os.path.join(DEST_DIR, PROJECT_NAME, PROJECT_NAME + "map.csv"), index=False
        )
    mapdf = mapdf.reset_index(drop=True)
    os.chdir(pwd)
    return json.dumps(mapdf.to_dict("list"))


def main():
    """
    Main function used in script, primarily used as a handle
    to get the output into stdout.
    """
    # There are no args, but parse them just so help works
    print(process_files_json(), end="")
    return None


def add_suffix_to_filename(filename, suffix):
    """
    Utility for modify filename strings with _suffix modifier, e. g.

    >>>print(add_suffix_to_filename("this_file.json", "_processed"))
    >>>"this_file_processed.json"

    Args:
        filename (str): filename to be modified.
        suffix (str): suffix to be added before final '.'.

    Returns:
        str: modified string with suffix.

    """
    name, ext = os.path.splitext(filename)
    return "".join([name, suffix, ext])


def scrub_underscore_suffix(filename):
    """
    Creates a new string from old filename with suffix e. g. "_processed"
    in "this_file_processed.json".

    Args:
        filename (str): filename with some suffix to be scrubbed.

    Returns:
        str: filename with suffix scrubbed.

    """
    scrubbed = re.sub(r"_[^_]+\.", ".", filename)
    return scrubbed


if __name__ == "__main__":
    main()



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and scripts for generating feature objects from structured
battery cycling data, to be used as inputs for machine learning
early prediction models.

Usage:
    featurize [INPUT_JSON]

Options:
    -h --help        Show this screen
    --version        Show version


The `featurize` script will generate features according to the methods
contained in beep.featurize.  It places output files corresponding to
features in `/data-share/features/`.

The input json must contain the following fields

* `file_list` - a list of processed cycler runs for which to generate features

The output json file will contain the following:

* `file_list` - a list of filenames corresponding to the locations of the features

Example:
```angular2
$ featurize '{"invalid_file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv"],
    "file_list": ["/data-share/structure/FastCharge_2_CH29_structure.json"]}'
{"file_list": ["/data-share/features/FastCharge_2_CH29_full_model_features.json"]}
```
"""

import os
import pandas as pd
from abc import ABCMeta, abstractmethod
from monty.json import MSONable
from monty.serialization import loadfn

from beep.collate import scrub_underscore_suffix, add_suffix_to_filename
from beep import MODULE_DIR

FEATURE_HYPERPARAMS = loadfn(
    os.path.join(MODULE_DIR, "features/feature_hyperparameters.yaml")
)

s = {"service": "DataAnalyzer"}


class BeepFeatures(MSONable, metaclass=ABCMeta):
    """
    Class corresponding to feature baseline feature object.

    Attributes:
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    class_feature_name = "Base"

    def __init__(self, name, X, metadata):
        """
        Invokes BeepFeatures object

        Args:
            name (str): predictor object name.
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the conditions, data
                and code used to produce features

        """
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def from_run(
        cls, input_filename, feature_dir, processed_cycler_run,
            params_dict=None, parameters_path="data-share/raw/parameters"
    ):
        """
        This method contains the workflow for the creation of the feature class
        Since the workflow should be the same for all of the feature classed this
        method should not be overridden in any of the derived classes. If the class
        can be created (feature generation succeeds, etc.) then the class is returned.
        Otherwise the return value is False
        Args:
            input_filename (str): path to the input data from processed cycler run
            feature_dir (str): path to the base directory for the feature sets.
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            (beep.featurize.BeepFeatures): class object for the feature set
        """

        if cls.validate_data(processed_cycler_run, params_dict):
            output_filename = cls.get_feature_object_name_and_path(
                input_filename, feature_dir
            )
            feature_object = cls.features_from_processed_cycler_run(
                processed_cycler_run, params_dict, parameters_path=parameters_path
            )
            metadata = cls.metadata_from_processed_cycler_run(
                processed_cycler_run, params_dict
            )
            return cls(output_filename, feature_object, metadata)
        else:
            return False

    @classmethod
    @abstractmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        Method for validation of input data, e.g. processed_cycler_runs

        Args:
            processed_cycler_run (ProcessedCyclerRun): processed_cycler_run
                to validate
            params_dict (dict): parameter dictionary for validation

        Returns:
            (bool): boolean for whether data is validated

        """
        raise NotImplementedError

    @classmethod
    def get_feature_object_name_and_path(cls, input_path, feature_dir):
        """
        This function determines how to name the object for a specific feature class
        and creates the full path to save the object. This full path is also used as
        the feature name attribute
        Args:
            input_path (str): path to the input data from processed cycler run
            feature_dir (str): path to the base directory for the feature sets.
        Returns:
            str: the full path (including filename) to use for saving the feature
                object
        """
        new_filename = os.path.basename(input_path)
        new_filename = scrub_underscore_suffix(new_filename)

        # Append model_name along with "features" to demarcate
        # different models when saving the feature vectors.
        new_filename = add_suffix_to_filename(
            new_filename, "_features" + "_" + cls.class_feature_name
        )
        if not os.path.isdir(os.path.join(feature_dir, cls.class_feature_name)):
            os.makedirs(os.path.join(feature_dir, cls.class_feature_name))
        feature_path = os.path.join(feature_dir, cls.class_feature_name, new_filename)
        feature_path = os.path.abspath(feature_path)
        return feature_path

    @classmethod
    @abstractmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        raise NotImplementedError

    @classmethod
    def metadata_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None):
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]
        metadata = {
            "barcode": processed_cycler_run.metadata.barcode,
            "protocol": processed_cycler_run.metadata.protocol,
            "channel_id": processed_cycler_run.metadata.channel_id,
            "parameters": params_dict,
        }
        return metadata

    def as_dict(self):
        """
        Method for dictionary serialization
        Returns:
            dict: corresponding to dictionary for serialization
        """
        obj = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "name": self.name,
            "X": self.X.to_dict("list"),
            "metadata": self.metadata,
        }
        return obj

    @classmethod
    def from_dict(cls, d):
        """MSONable deserialization method"""
        d["X"] = pd.DataFrame(d["X"])
        return cls(**d)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for generating maccor procedure files from
input parameters and procedure templates
"""

import re
from collections import OrderedDict
from copy import deepcopy

from beep.utils import DashOrderedDict

s = {"service": "ProtocolGenerator"}


class Settings(DashOrderedDict):
    """
    Settings file object. Provides factory methods
    to read a Biologic-type settings file and invoke
    from templates for specific experimental
     parameters

    """

    @classmethod
    def from_file(
        cls, filename, encoding="ISO-8859-1", column_width=20, step_entry_length=63
    ):
        """
        Settings file ingestion. Invokes Settings object
        from standard Biologic *.mps file.

        Args:
            filename (str): settings file.
            encoding (str): file encoding to use when reading the file
            column_width (int): number of characters per step column
            step_entry_length (int): This is the number of lines in the steps. It is based on the format seen
                in one file, and might need to be changed or dynamically determined upon file read

        Returns:
            (Settings): Ordered dictionary with keys corresponding to options or
                control variables. Section headers are nested dicts or lists
                within the dict.
        """
        with open(filename, "rb") as f:
            text = f.read()
            text = text.decode(encoding)

        return cls.mps_text_to_schedule_dict(text, column_width, step_entry_length)

    @classmethod
    def mps_text_to_schedule_dict(cls, text, column_width=20, step_entry_length=63):
        obj = cls()
        split_text = re.split(r"\r?\n|\r\n?", text)

        extra_lines = list(range(len(split_text)))
        technique_lines = [
            indx for indx, val in enumerate(split_text) if "Technique" in val
        ]
        technique_pos = []
        for technique_start_line in technique_lines:
            technique_num = split_text[technique_start_line].split(":")[-1].strip()
            start = (
                technique_start_line + 1
            )  # This +1 offset is based on the format seen in one file
            end = start + step_entry_length
            technique_pos.append((technique_num, start, end))
            lines_to_parse = [
                i for i in extra_lines if i not in list(range(start, end))
            ]

        section = "Metadata"
        metadata = []
        for line_num in lines_to_parse:
            if ":" in split_text[line_num]:
                if "Technique" in split_text[line_num]:
                    metadata.append(
                        [
                            "_".join(split_text[line_num].split(" : ", 1)),
                            split_text[line_num].split(":", 1)[-1].strip(),
                        ]
                    )
                else:
                    metadata.append(split_text[line_num].split(" : ", 1))
            elif split_text[line_num] == "":
                metadata.append(["line{}".format(line_num), "blank"])
            else:
                metadata.append([split_text[line_num], None])
        meta = OrderedDict(metadata)
        obj.set("{}".format(section), meta)

        section = "Technique"
        for technique in technique_pos:
            technique_num = technique[0]
            start = technique[1]
            end = technique[2]
            obj.set(
                "{}.{}.{}".format(section, technique_num, "Type"), split_text[start]
            )
            start = start + 1
            number_of_columns = (
                max([len(l) for l in split_text[start:end]]) // column_width
            )
            technique_steps = []
            for line in split_text[start:end]:
                steps_values = []
                for col in range(number_of_columns):
                    steps_values.append(
                        line[col * column_width: (col + 1) * column_width].strip()
                    )
                technique_steps.append(steps_values)
            step_matrix = list(zip(*technique_steps))
            step_headers = step_matrix[0]

            for step_number in range(1, number_of_columns):
                step = OrderedDict(zip(step_headers, step_matrix[step_number]))
                obj.set(
                    "{}.{}.{}.{}".format(
                        section, technique_num, "Step", str(step_number)
                    ),
                    step,
                )

        return obj

    def to_file(self, filename, encoding="ISO-8859-1", column_width=20, linesep="\r\n"):
        """
        Write DashOrderedDict to a settings file in the Biologic format with a *.mps extension.

        Args:
            filename (str): output name for settings file, full path
            encoding (str): file encoding to use when reading the file
            column_width (int): number of characters per step column
            linesep (str): characters to use for line separation, usually CRLF for Windows based machines

        Returns:
            (None): Writes out data to filename
        """
        data = deepcopy(self)
        blocks = []
        meta_data_keys = list(data["Metadata"].keys())
        for indx, meta_key in enumerate(meta_data_keys):
            if "Technique_" in meta_key:
                blocks.append(
                    " : ".join([meta_key.split("_")[0], data["Metadata"][meta_key]])
                )
                tq_number = data["Metadata"][meta_key]

                blocks.append(data["Technique"][tq_number]["Type"])
                technique_keys = list(data["Technique"][tq_number]["Step"]["1"].keys())
                for tq_key in technique_keys:
                    line = tq_key.ljust(column_width)
                    for step in data["Technique"][tq_number]["Step"].keys():
                        var = str(data["Technique"][tq_number]["Step"][step][tq_key])
                        line = line + var.ljust(column_width)
                    blocks.append(line)
                continue
            elif data["Metadata"][meta_key] is None:
                line = meta_key
            elif data["Metadata"][meta_key] == "blank":
                line = ""
            elif data["Metadata"][meta_key] is not None:
                line = " : ".join([meta_key, data["Metadata"][meta_key]])
            blocks.append(line)
            data.unset(data["Metadata"][meta_key])

        contents = linesep.join(blocks)

        with open(filename, "wb") as f:
            f.write(contents.encode(encoding))

    def formation_protocol_bcs(self, params):
        """
       Modify a settings file in the Biologic format with a *.mps extension.

        Args:
            params (pandas.DataFrame): Single row of the csv containing the parameters for the project

        Returns:
            (self): returns the parameterized protocol
        """
        assert self.get_path("Metadata.Cycle Definition") == "Charge/Discharge alternance"

        # Initial charge
        assert self.get_path("Technique.1.Step.2.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.2.ctrl1_val_unit") == "A"
        value = float(round(params["initial_charge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.2.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.2.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.2.lim1_value_unit") == "V"
        value = float(round(params["initial_charge_voltage_1"], 3))
        self.set("Technique.1.Step.2.lim1_value", value)
        assert self.get_path("Technique.1.Step.3.ctrl_type") == "CV"
        assert self.get_path("Technique.1.Step.3.ctrl1_val_unit") == "V"
        self.set("Technique.1.Step.3.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.3.lim1_type") == "Time"
        assert self.get_path("Technique.1.Step.3.lim1_value_unit") == "mn"
        value = float(round(params["initial_charge_cvhold_1"], 3))
        self.set("Technique.1.Step.3.lim1_value", value)

        assert self.get_path("Technique.1.Step.4.lim1_type") == "Time"
        assert self.get_path("Technique.1.Step.4.lim1_value_unit") == "mn"
        value = float(round(params["initial_charge_rest_1"], 3))
        self.set("Technique.1.Step.4.lim1_value", value)

        assert self.get_path("Technique.1.Step.5.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.5.ctrl1_val_unit") == "A"
        value = float(round(params["initial_charge_current_2"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.5.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.5.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.5.lim1_value_unit") == "V"
        value = float(round(params["initial_charge_voltage_2"], 3))
        self.set("Technique.1.Step.5.lim1_value", value)
        assert self.get_path("Technique.1.Step.6.ctrl_type") == "CV"
        assert self.get_path("Technique.1.Step.6.ctrl1_val_unit") == "V"
        self.set("Technique.1.Step.6.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.6.lim1_type") == "Time"
        assert self.get_path("Technique.1.Step.6.lim1_value_unit") == "mn"
        value = float(round(params["initial_charge_cvhold_2"], 3))
        self.set("Technique.1.Step.6.lim1_value", value)

        assert self.get_path("Technique.1.Step.7.lim1_type") == "Time"
        assert self.get_path("Technique.1.Step.7.lim1_value_unit") == "mn"
        value = float(round(params["initial_charge_rest_2"], 3))
        self.set("Technique.1.Step.7.lim1_value", value)

        # Initial discharge
        assert self.get_path("Technique.1.Step.8.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.8.ctrl1_val_unit") == "A"
        value = float(round(params["initial_discharge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.8.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.8.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.8.lim1_value_unit") == "V"
        value = float(round(params["initial_discharge_voltage_1"], 3))
        self.set("Technique.1.Step.8.lim1_value", value)

        # Mid cycle
        assert self.get_path("Technique.1.Step.9.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.9.ctrl1_val_unit") == "A"
        value = float(round(params["mid_charge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.9.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.9.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.9.lim1_value_unit") == "V"
        value = float(round(params["mid_charge_voltage_1"], 3))
        self.set("Technique.1.Step.9.lim1_value", value)

        assert self.get_path("Technique.1.Step.10.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.10.ctrl1_val_unit") == "A"
        value = float(round(params["mid_discharge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.10.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.10.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.10.lim1_value_unit") == "V"
        value = float(round(params["mid_discharge_voltage_1"], 3))
        self.set("Technique.1.Step.10.lim1_value", value)

        assert self.get_path("Technique.1.Step.11.ctrl_type") == "Loop"
        assert self.get_path("Technique.1.Step.11.ctrl_seq") == "8"
        value = int(params["mid_cycle_reps"] - 1)
        self.set("Technique.1.Step.11.ctrl_repeat", value)

        # Final cycle
        assert self.get_path("Technique.1.Step.12.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.12.ctrl1_val_unit") == "A"
        value = float(round(params["final_charge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.12.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.12.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.12.lim1_value_unit") == "V"
        value = float(round(params["final_charge_voltage_1"], 3))
        self.set("Technique.1.Step.12.lim1_value", value)

        assert self.get_path("Technique.1.Step.13.ctrl_type") == "CC"
        assert self.get_path("Technique.1.Step.13.ctrl1_val_unit") == "A"
        value = float(round(params["final_discharge_current_1"] * params['capacity_nominal'], 3))
        self.set("Technique.1.Step.13.ctrl1_val", value)

        assert self.get_path("Technique.1.Step.13.lim1_type") == "Ecell"
        assert self.get_path("Technique.1.Step.13.lim1_value_unit") == "V"
        value = float(round(params["final_discharge_voltage_1"], 3))
        self.set("Technique.1.Step.13.lim1_value", value)
        assert self.get_path("Technique.1.Step.14.ctrl_type") == "CV"
        assert self.get_path("Technique.1.Step.14.ctrl1_val_unit") == "V"
        self.set("Technique.1.Step.14.ctrl1_val", value)
        assert self.get_path("Technique.1.Step.14.lim1_type") == "Time"
        assert self.get_path("Technique.1.Step.14.lim1_value_unit") == "mn"
        value = float(round(params["final_discharge_cvhold_1"], 3))
        self.set("Technique.1.Step.14.lim1_value", value)

        return self



"""
Configuration file for various runtime environments for BEEP-EP.


"""

config = {
    "local": {
        "logging": {"container": "Testing", "streams": ["file"]},
        "kinesis": {"stream": "local/beep/eventstream"},
    },
    "dev": {
        "logging": {"container": "Testing", "streams": ["file"]},
        "kinesis": {"stream": "local/beep/eventstream"},
    },
    "test": {
        "logging": {"container": "Testing", "streams": ["file"]},
        "kinesis": {"stream": "local/beep/eventstream"},
    },
    "stage": {
        "logging": {"container": "BEEP_EP", "streams": ["CloudWatch", "stdout"]},
        "kinesis": {"stream": "stage/beep/eventstream/stage"},
    },
    "prod": {},
}



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and scripts for generating feature objects from structured
battery cycling data, to be used as inputs for machine learning
early prediction models.

Usage:
    featurize [INPUT_JSON]

Options:
    -h --help        Show this screen
    --version        Show version


The `featurize` script will generate features according to the methods
contained in beep.featurize.  It places output files corresponding to
features in `/data-share/features/`.

The input json must contain the following fields

* `file_list` - a list of processed cycler runs for which to generate features

The output json file will contain the following:

* `file_list` - a list of filenames corresponding to the locations of the features

Example:
```angular2
$ featurize '{"invalid_file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv"],
    "file_list": ["/data-share/structure/FastCharge_2_CH29_structure.json"]}'
{"file_list": ["/data-share/features/FastCharge_2_CH29_full_model_features.json"]}
```
"""

import os
import json
import numpy as np
import pandas as pd
from docopt import docopt
from monty.json import MSONable
from monty.serialization import loadfn, dumpfn
from scipy.stats import skew, kurtosis

from beep.structure.cli import auto_load_processed
from beep.utils import WorkflowOutputs
from beep.features import featurizer_helpers, intracell_losses
from beep.features.base import BeepFeatures
from beep import logger, __version__

MODULE_DIR = os.path.dirname(__file__)
FEATURE_HYPERPARAMS = loadfn(
    os.path.join(MODULE_DIR, "features/feature_hyperparameters.yaml")
)

s = {"service": "DataAnalyzer"}


class HPPCResistanceVoltageFeatures(BeepFeatures):
    """
    Object corresponding to resistance, voltage and diffusion related
    features generated from hybrid pulse power characterization cycles.
    Includes constructors to create the features, object names and metadata.

    Attributes:
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features

    """

    # Class name for the feature object
    class_feature_name = "HPPCResistanceVoltageFeatures"

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations

        Returns:
            (bool): True/False indication of ability to proceed with feature generation
        """
        conditions = []
        if not hasattr(processed_cycler_run, "diagnostic_summary") & hasattr(
            processed_cycler_run, "diagnostic_data"
        ):
            return False
        if processed_cycler_run.diagnostic_summary is None:
            return False
        elif processed_cycler_run.diagnostic_summary.empty:
            return False
        else:
            conditions.append(
                any(
                    [
                        "hppc" in x
                        for x in processed_cycler_run.diagnostic_summary.cycle_type.unique()
                    ]
                )
            )

        return all(conditions)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        This method calculates features based on voltage, diffusion and resistance changes in hppc cycles.

        Note: Inside this function it calls function get_dr_df, but if the cell does not state of charge from 20% to
        10%, the function will fail, and throw you error messages. This will only happen after cycle 37 and on fast
        charging cells. Also, this function calls function get_v_diff, which takes in an argument soc_window, if you
        want more correlation, you should go for low state of charge, which corresponds to soc_window = 8. However,
        like the resistance feature, at cycle 142 and beyond, soc_window = 8 might fail on fast charged cells. For
        lower soc_window values, smaller than or equal to 7, this should not be a problem, but it will not give you
        high correlations.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.


        Returns:
            dataframe of features based on voltage and resistance changes over a SOC window in hppc cycles
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        # Filter out low cycle numbers at the end of the test, corresponding to the "final" diagnostic
        processed_cycler_run.diagnostic_data = processed_cycler_run.diagnostic_data[
                ~((processed_cycler_run.diagnostic_data.test_time > params_dict['test_time_filter_sec']) &
                  (processed_cycler_run.diagnostic_data.cycle_index < params_dict['cycle_index_filter']))
            ]
        processed_cycler_run.diagnostic_data = processed_cycler_run.diagnostic_data.groupby(
            ["cycle_index", "step_index", "step_index_counter"]
        ).filter(lambda x: ~x["test_time"].isnull().all())

        # diffusion features
        diffusion_features = featurizer_helpers.get_diffusion_features(
            processed_cycler_run, params_dict["diag_pos"]
        )

        hppc_r = pd.DataFrame()
        # the 9 by 6 dataframe
        df_dr = featurizer_helpers.get_dr_df(
            processed_cycler_run, params_dict["diag_pos"]
        )
        # transform this dataframe to be 1 by 54
        columns = df_dr.columns
        for column in columns:
            for r in range(len(df_dr[column])):
                name = column + str(r)
                hppc_r[name] = [df_dr[column][r]]

        # the variance of ocv features
        hppc_ocv = featurizer_helpers.get_hppc_ocv(
            processed_cycler_run, params_dict["diag_pos"]
        )

        # the v_diff features
        v_diff = featurizer_helpers.get_v_diff(
            processed_cycler_run, params_dict["diag_pos"], params_dict["soc_window"]
        )

        # merge everything together as a final result dataframe
        return pd.concat([hppc_r, hppc_ocv, v_diff, diffusion_features], axis=1)


class CycleSummaryStats(BeepFeatures):
    """
    Object corresponding to summary statistics from
    regular cycles

    name (str): predictor object name.
    X (pandas.DataFrame): features in DataFrame format.
    metadata (dict): information about the conditions, data
        and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "CycleSummaryStats"

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        # TODO: not sure this is necessary
        # Check for data in each of the selected cycles
        index_1, index_2 = params_dict['cycle_comp_num']
        cycle_1 = processed_cycler_run.structured_data[
            processed_cycler_run.structured_data.cycle_index == index_1]
        cycle_2 = processed_cycler_run.structured_data[
            processed_cycler_run.structured_data.cycle_index == index_2]
        if len(cycle_1) == 0 or len(cycle_2) == 0:
            return False

        # TODO: check whether this is good
        # Check for relevant data
        required_columns = ['charge_capacity',
                            'discharge_capacity',
                            'charge_energy',
                            'discharge_energy',
                            ]
        pcycler_run_columns = processed_cycler_run.structured_data.columns
        if not all([column in pcycler_run_columns for column in required_columns]):
            return False

        return True

    SUMMARY_STATISTIC_NAMES = ["var", "min", "mean", "skew", "kurtosis", "abs", "square"]

    @staticmethod
    def get_summary_statistics(array):
        """
        Static method for getting values corresponding
        to standard 7 operations that many beep features
        use, i.e. log of absolute value of each of
        variance, min, mean, skew, kurtosis, the sum of
        the absolute values and the sum of squares

        Args:
            array (list, np.ndarray): array of values to get
                standard operation values for, e.g. cycle
                discharging capacity, QcDiff, etc.

        Returns:
            [float]: list of features

        """
        return np.array([
            np.log10(np.absolute(np.var(array))),
            np.log10(np.absolute(min(array))),
            np.log10(np.absolute(np.mean(array))),
            np.log10(np.absolute(skew(array))),
            np.log10(np.absolute(kurtosis(array, fisher=False, bias=False))),
            np.log10(np.sum(np.absolute(array))),
            np.log10(np.sum(np.square(array)))
        ])

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        Generate features listed in early prediction manuscript using both diagnostic and regular cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            X (pd.DataFrame): Dataframe containing the feature
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        # TODO: extend this dataframe and uncomment energy features when
        #   structuring is refactored
        X = pd.DataFrame(np.zeros((1, 28)))

        reg_cycle_comp_num = params_dict.get("cycle_comp_num")
        cycle_comp_1 = processed_cycler_run.structured_data[
            processed_cycler_run.structured_data.cycle_index == reg_cycle_comp_num[1]
            ]
        cycle_comp_0 = processed_cycler_run.structured_data[
            processed_cycler_run.structured_data.cycle_index == reg_cycle_comp_num[0]
            ]
        Qc100_1 = cycle_comp_1[cycle_comp_1.step_type == "charge"].charge_capacity
        Qc10_1 = cycle_comp_0[cycle_comp_0.step_type == "charge"].charge_capacity
        QcDiff = Qc100_1.values - Qc10_1.values
        QcDiff = QcDiff[~np.isnan(QcDiff)]

        X.loc[0, 0:6] = cls.get_summary_statistics(QcDiff)

        Qd100_1 = cycle_comp_1[cycle_comp_1.step_type == "discharge"].discharge_capacity
        Qd10_1 = cycle_comp_0[cycle_comp_0.step_type == "discharge"].discharge_capacity
        QdDiff = Qd100_1.values - Qd10_1.values
        QdDiff = QdDiff[~np.isnan(QdDiff)]

        X.loc[0, 7:13] = cls.get_summary_statistics(QdDiff)

        # # Charging Energy features
        Ec100_1 = cycle_comp_1[cycle_comp_1.step_type == "charge"].charge_energy
        Ec10_1 = cycle_comp_0[cycle_comp_0.step_type == "charge"].charge_energy
        EcDiff = Ec100_1.values - Ec10_1.values
        EcDiff = EcDiff[~np.isnan(EcDiff)]

        X.loc[0, 14:20] = cls.get_summary_statistics(EcDiff)

        # # Discharging Energy features
        Ed100_1 = cycle_comp_1[cycle_comp_1.step_type == "charge"].discharge_energy
        Ed10_1 = cycle_comp_0[cycle_comp_0.step_type == "charge"].discharge_energy
        EdDiff = Ed100_1.values - Ed10_1.values
        EdDiff = EdDiff[~np.isnan(EdDiff)]

        X.loc[0, 21:27] = cls.get_summary_statistics(EdDiff)

        quantities = [
            "charging_capacity",
            "discharging_capacity",
            "charging_energy",
            "discharging_energy",
        ]

        X.columns = [y + "_" + x for x in quantities for y in cls.SUMMARY_STATISTIC_NAMES]

        return X


class DiagnosticSummaryStats(CycleSummaryStats):
    """
    Object corresponding to summary statistics from a diagnostic cycle of
    specific type. Includes constructors to create the features, object names
    and metadata attributes in the object.  Inherits from RegularCycleSummaryStats
    to reuse standard feature generation

    name (str): predictor object name.
    X (pandas.DataFrame): features in DataFrame format.
    metadata (dict): information about the conditions, data
        and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "DiagnosticSummaryStats"

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations

        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]
        conditions = []
        if not hasattr(processed_cycler_run, "diagnostic_summary") & hasattr(
            processed_cycler_run, "diagnostic_data"
        ):
            return False
        if processed_cycler_run.diagnostic_summary is None:
            return False
        elif processed_cycler_run.diagnostic_summary.empty:
            return False
        else:
            df = processed_cycler_run.diagnostic_summary
            df = df[df.cycle_type == params_dict["diagnostic_cycle_type"]]
            conditions.append(
                df.cycle_index.nunique() >= max(params_dict["diag_pos_list"]) + 1
            )

        return all(conditions)

    @classmethod
    def get_summary_diff(cls, processed_cycler_run,
                         pos=None,
                         cycle_types=None,
                         metrics=None):
        """
        Helper function to calculate difference between summary values in the diagnostic cycles

                Args:
                    processed_cycler_run (beep.structure.ProcessedCyclerRun)
                    pos (list): position of the diagnostics to use in the calculation
                    cycle_types (list): calculate difference for these diagnostic types
                    metrics (str): Calculate difference for these metrics

                Returns:
                    values (list): List of difference values to insert into the dataframe
                    names (list): List of column headers to use in the creation of the dataframe
                """
        if pos is None:
            pos = [0, 1]
        if cycle_types is None:
            cycle_types = ["rpt_0.2C", "rpt_1C", "rpt_2C"]
        if metrics is None:
            metrics = ["discharge_capacity", "discharge_energy", "charge_capacity", "charge_energy"]

        values = []
        names = []
        for cycle_type in cycle_types:
            diag_type_summary = processed_cycler_run.diagnostic_summary[
                processed_cycler_run.diagnostic_summary.cycle_type == cycle_type]
            for metric in metrics:
                diff = (diag_type_summary.iloc[pos[1]][metric] - diag_type_summary.iloc[pos[0]][metric]) \
                       / diag_type_summary.iloc[pos[0]][metric]
                values.append(diff)
                names.append("diag_sum_diff_" + str(pos[0]) + "_" + str(pos[1]) + "_" + cycle_type + metric)
        return values, names

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        Generate features listed in early prediction manuscript using both diagnostic and regular cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            X (pd.DataFrame): Dataframe containing the feature
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        # Filter out "final" diagnostic cycles that have been appended to the end of the file with the wrong
        # cycle number(test time is monotonic)
        processed_cycler_run.diagnostic_data = processed_cycler_run.diagnostic_data[
            ~((processed_cycler_run.diagnostic_data.test_time > params_dict['test_time_filter_sec']) &
              (processed_cycler_run.diagnostic_data.cycle_index < params_dict['cycle_index_filter']))
        ]
        processed_cycler_run.diagnostic_data = processed_cycler_run.diagnostic_data.groupby(
            ["cycle_index", "step_index", "step_index_counter"]
        ).filter(lambda x: ~x["test_time"].isnull().all())

        diag_intrp = processed_cycler_run.diagnostic_data

        X = pd.DataFrame(np.zeros((1, 54)))

        # Calculate the cycles and the steps for the selected diagnostics
        cycles = diag_intrp.cycle_index[diag_intrp.cycle_type ==
                                        params_dict["diagnostic_cycle_type"]].unique()
        step_dict_0 = featurizer_helpers.get_step_index(processed_cycler_run,
                                                        cycle_type=params_dict["diagnostic_cycle_type"],
                                                        diag_pos=params_dict["diag_pos_list"][0])
        step_dict_1 = featurizer_helpers.get_step_index(processed_cycler_run,
                                                        cycle_type=params_dict["diagnostic_cycle_type"],
                                                        diag_pos=params_dict["diag_pos_list"][1])

        # Create masks for each position in the data
        mask_pos_0_charge = ((diag_intrp.cycle_index == cycles[params_dict["diag_pos_list"][0]]) &
                             (diag_intrp.step_index == step_dict_0[params_dict["diagnostic_cycle_type"] + '_charge']))
        mask_pos_1_charge = ((diag_intrp.cycle_index == cycles[params_dict["diag_pos_list"][1]]) &
                             (diag_intrp.step_index == step_dict_1[params_dict["diagnostic_cycle_type"] + '_charge']))
        mask_pos_0_discharge = ((diag_intrp.cycle_index == cycles[params_dict["diag_pos_list"][0]]) &
                                (diag_intrp.step_index ==
                                 step_dict_0[params_dict["diagnostic_cycle_type"] + '_discharge']))
        mask_pos_1_discharge = ((diag_intrp.cycle_index == cycles[params_dict["diag_pos_list"][1]]) &
                                (diag_intrp.step_index ==
                                 step_dict_1[params_dict["diagnostic_cycle_type"] + '_discharge']))

        # Charging Capacity features
        Qc_1 = diag_intrp.charge_capacity[mask_pos_1_charge]
        Qc_0 = diag_intrp.charge_capacity[mask_pos_0_charge]
        QcDiff = Qc_1.values - Qc_0.values
        QcDiff = QcDiff[~np.isnan(QcDiff)]

        X.loc[0, 0:6] = cls.get_summary_statistics(QcDiff)

        # Discharging Capacity features
        Qd_1 = diag_intrp.discharge_capacity[mask_pos_1_discharge]
        Qd_0 = diag_intrp.discharge_capacity[mask_pos_0_discharge]
        QdDiff = Qd_1.values - Qd_0.values
        QdDiff = QdDiff[~np.isnan(QdDiff)]

        X.loc[0, 7:13] = cls.get_summary_statistics(QdDiff)

        # Charging Energy features
        Ec_1 = diag_intrp.charge_energy[mask_pos_1_charge]
        Ec_0 = diag_intrp.charge_energy[mask_pos_0_charge]
        EcDiff = Ec_1.values - Ec_0.values
        EcDiff = EcDiff[~np.isnan(EcDiff)]

        X.loc[0, 14:20] = cls.get_summary_statistics(EcDiff)

        # Discharging Energy features
        Ed_1 = diag_intrp.discharge_energy[mask_pos_1_discharge]
        Ed_0 = diag_intrp.discharge_energy[mask_pos_0_discharge]
        EdDiff = Ed_1.values - Ed_0.values
        EdDiff = EdDiff[~np.isnan(EdDiff)]

        X.loc[0, 21:27] = cls.get_summary_statistics(EdDiff)

        # Charging dQdV features
        dQdVc_1 = diag_intrp.charge_dQdV[mask_pos_1_charge]
        dQdVc_0 = diag_intrp.charge_dQdV[mask_pos_0_charge]
        dQdVcDiff = dQdVc_1.values - dQdVc_0.values
        dQdVcDiff = dQdVcDiff[~np.isnan(dQdVcDiff)]

        X.loc[0, 28:34] = cls.get_summary_statistics(dQdVcDiff)

        # Discharging Capacity features
        dQdVd_1 = diag_intrp.discharge_dQdV[mask_pos_1_discharge]
        dQdVd_0 = diag_intrp.discharge_dQdV[mask_pos_0_discharge]
        dQdVdDiff = dQdVd_1.values - dQdVd_0.values
        dQdVdDiff = dQdVdDiff[~np.isnan(dQdVdDiff)]

        X.loc[0, 35:41] = cls.get_summary_statistics(dQdVdDiff)

        X.loc[0, 42:53], names = cls.get_summary_diff(processed_cycler_run, params_dict["diag_pos_list"])

        operations = ["var", "min", "mean", "skew", "kurtosis", "abs", "square"]
        quantities = [
            "charging_capacity",
            "discharging_capacity",
            "charging_energy",
            "discharging_energy",
            "charging_dQdV",
            "discharging_dQdV",
        ]

        X.columns = [y + "_" + x for x in quantities for y in operations] + names
        return X


class DeltaQFastCharge(BeepFeatures):
    """
    Object corresponding to feature object. Includes constructors
    to create the features, object names and metadata attributes in the
    object
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "DeltaQFastCharge"

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """

        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        conditions = list()

        conditions.append(
            processed_cycler_run.structured_summary.index.max()
            > params_dict["final_pred_cycle"]
        )
        conditions.append(
            processed_cycler_run.structured_summary.index.min()
            <= params_dict["init_pred_cycle"]
        )
        conditions.append("cycle_index" in processed_cycler_run.structured_summary.columns)
        conditions.append("cycle_index" in processed_cycler_run.structured_data.columns)

        return all(conditions)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        Generate features listed in early prediction manuscript, primarily related to the
        so called delta Q feature
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: features indicative of degradation, derived from the input data
        """

        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        assert params_dict["mid_pred_cycle"] > 10  # Sufficient cycles for analysis
        assert (
            params_dict["final_pred_cycle"] > params_dict["mid_pred_cycle"]
        )  # Must have final_pred_cycle > mid_pred_cycle

        i_final = params_dict["final_pred_cycle"] - 1  # python indexing
        i_mid = params_dict["mid_pred_cycle"] - 1

        summary = processed_cycler_run.structured_summary
        params_dict[
            "n_nominal_cycles"
        ] = 40  # For nominal capacity, use median discharge capacity of first n cycles

        if "step_type" in processed_cycler_run.structured_data.columns:
            interpolated_df = processed_cycler_run.structured_data[
                processed_cycler_run.structured_data.step_type == "discharge"
            ]
        else:
            interpolated_df = processed_cycler_run.structured_data
        X = pd.DataFrame(np.zeros((1, 20)))
        labels = []
        # Discharge capacity, cycle 2 = Q(n=2)
        X[0] = summary.discharge_capacity.iloc[1]
        labels.append("discharge_capacity_cycle_2")

        # Max discharge capacity - discharge capacity, cycle 2 = max_n(Q(n)) - Q(n=2)
        X[1] = max(
            summary.discharge_capacity.iloc[np.arange(i_final + 1)]
            - summary.discharge_capacity.iloc[1]
        )
        labels.append("max_discharge_capacity_difference")

        # Discharge capacity, cycle 100 = Q(n=100)
        X[2] = summary.discharge_capacity.iloc[i_final]
        labels.append("discharge_capacity_cycle_100")

        # Feature representing time-temperature integral over cycles 2 to 100
        X[3] = np.nansum(summary.time_temperature_integrated.iloc[np.arange(i_final + 1)])
        labels.append("integrated_time_temperature_cycles_1:100")

        # Mean of charge times of first 5 cycles
        X[4] = np.nanmean(summary.charge_duration.iloc[1:6])
        labels.append("charge_time_cycles_1:5")

        # Descriptors based on capacity loss between cycles 10 and 100.
        Qd_final = interpolated_df.discharge_capacity[
            interpolated_df.cycle_index == i_final
        ]
        Qd_10 = interpolated_df.discharge_capacity[interpolated_df.cycle_index == 9]

        Qd_diff = Qd_final.values - Qd_10.values

        # If DeltaQ(V) is not an empty array, compute summary stats, else initialize with np.nan
        # Cells discharged rapidly over a narrow voltage window run into have no interpolated discharge steps
        if len(Qd_diff):
            X[5] = np.log10(np.abs(np.nanmin(Qd_diff)))  # Minimum
            X[6] = np.log10(np.abs(np.nanmean(Qd_diff)))  # Mean
            X[7] = np.log10(np.abs(np.nanvar(Qd_diff)))  # Variance
            X[8] = np.log10(np.abs(skew(Qd_diff)))  # Skewness
            X[9] = np.log10(np.abs(kurtosis(Qd_diff)))  # Kurtosis
            X[10] = np.log10(np.abs(Qd_diff[0]))  # First difference
        else:
            X[5:11] = np.nan

        labels.append("abs_min_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_mean_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_variance_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_skew_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_kurtosis_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_first_discharge_capacity_difference_cycles_2:100")

        X[11] = np.max(summary.temperature_maximum.iloc[list(range(1, i_final + 1))])  # Max T
        labels.append("max_temperature_cycles_1:100")

        X[12] = np.min(summary.temperature_minimum.iloc[list(range(1, i_final + 1))])  # Min T
        labels.append("min_temperature_cycles_1:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 2 to 100

        X[13], X[14] = np.polyfit(
            list(range(1, i_final + 1)),
            summary.discharge_capacity.iloc[list(range(1, i_final + 1))],
            1,
        )

        labels.append("slope_discharge_capacity_cycle_number_2:100")
        labels.append("intercept_discharge_capacity_cycle_number_2:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 91 to 100
        X[15], X[16] = np.polyfit(
            list(range(i_mid, i_final + 1)),
            summary.discharge_capacity.iloc[list(range(i_mid, i_final + 1))],
            1,
        )
        labels.append("slope_discharge_capacity_cycle_number_91:100")
        labels.append("intercept_discharge_capacity_cycle_number_91:100")

        IR_trend = summary.dc_internal_resistance.iloc[list(range(1, i_final + 1))]
        if any(v == 0 for v in IR_trend):
            IR_trend[IR_trend == 0] = np.nan

        # Internal resistance minimum
        X[17] = np.nanmin(IR_trend)
        labels.append("min_internal_resistance_cycles_2:100")

        # Internal resistance at cycle 2
        X[18] = summary.dc_internal_resistance.iloc[1]
        labels.append("internal_resistance_cycle_2")

        # Internal resistance at cycle 100 - cycle 2
        X[19] = (
            summary.dc_internal_resistance.iloc[i_final] - summary.dc_internal_resistance.iloc[1]
        )
        labels.append("internal_resistance_difference_cycles_2:100")

        # Nominal capacity
        X[20] = np.median(
            summary.discharge_capacity.iloc[0: params_dict["n_nominal_cycles"]]
        )
        labels.append("nominal_capacity_by_median")

        X.columns = labels
        return X


class TrajectoryFastCharge(DeltaQFastCharge):
    """
    Object corresponding to cycle numbers at which the capacity drops below
     specific percentages of the initial capacity. Computed on the discharge
     portion of the regular fast charge cycles.


    """

    # Class name for the feature object
    class_feature_name = "TrajectoryFastCharge"

    def __init__(self, name, X, metadata):
        """
        Invokes a TrajectoryFastCharge object

        Args:
            name (str): predictor object name.
            X (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the conditions, data
                and code used to produce features

        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """

        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        conditions = []
        cap = processed_cycler_run.structured_summary.discharge_capacity
        conditions.append(cap.min() / cap.max() < params_dict["thresh_max_cap"])

        return all(conditions)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        Calculate the outcomes from the input data. In particular, the number of cycles
        where we expect to reach certain thresholds of capacity loss
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: cycles at which capacity/energy degradation exceeds thresholds
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]
        y = processed_cycler_run.capacities_to_cycles(
            params_dict["thresh_max_cap"],
            params_dict["thresh_min_cap"],
            params_dict["interval_cap"],
        )
        return y


class DiagnosticProperties(BeepFeatures):
    """
    This class stores fractional levels of degradation in discharge capacity and discharge energy
    relative to the first cycle at each diagnostic cycle, grouped by diagnostic cycle type.

        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """

    # Class name for the feature object
    class_feature_name = "DiagnosticProperties"

    def __init__(self, name, X, metadata):
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run, params_dict=None):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        if not hasattr(processed_cycler_run, "diagnostic_summary") & hasattr(
            processed_cycler_run, "diagnostic_data"
        ):
            return False
        if processed_cycler_run.diagnostic_summary is None:
            return False
        elif processed_cycler_run.diagnostic_summary.empty:
            return False
        else:
            return True

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run, params_dict=None,
                                           parameters_path="data-share/raw/parameters"):
        """
        Generates diagnostic-property features from processed cycler run, including values for n*x method
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: with "cycle_index", "fractional_metric", "x", "n", "cycle_type" and "metric" columns, rows
            for each diagnostic cycle of the cell
        """
        if params_dict is None:
            params_dict = FEATURE_HYPERPARAMS[cls.class_feature_name]

        cycle_types = processed_cycler_run.diagnostic_summary.cycle_type.unique()
        X = pd.DataFrame()
        for quantity in params_dict["quantities"]:
            for cycle_type in cycle_types:
                summary_diag_cycle_type = featurizer_helpers.get_fractional_quantity_remaining_nx(
                    processed_cycler_run, quantity, cycle_type, parameters_path=parameters_path
                )

                summary_diag_cycle_type.loc[:, "cycle_type"] = cycle_type
                summary_diag_cycle_type.loc[:, "metric"] = quantity
                X = X.append(summary_diag_cycle_type)

        return X


class DegradationPredictor(MSONable):
    """
    Object corresponding to feature matrix. Includes constructors
    to initialize the feature vectors.
    Attributes:
        name (str): predictor object name.
        X (pandas.DataFrame): data as records x features.
        y (pandas.DataFrame): targets.
        feature_labels (list): feature labels.
        predict_only (bool): True/False to specify predict/train mode.
        prediction_type (str): Type of regression - 'single' vs 'multi'.
        predicted_quantity (str): 'cycle' or 'capacity'.
        nominal_capacity (float):
    """

    def __init__(
        self,
        name,
        X,
        feature_labels=None,
        y=None,
        nominal_capacity=1.1,
        predict_only=False,
        predicted_quantity="cycle",
        prediction_type="multi",
    ):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            name (str): name of method for featurization.
            y (pandas.Dataframe or float): one or more outcomes.
            predict_only (bool): True/False to specify predict/train mode.
            predicted_quantity (str): 'cycle' or 'capacity'.
            prediction_type (str): Type of regression - 'single' vs 'multi'.
        """
        self.name = name
        self.X = X
        self.feature_labels = feature_labels
        self.predict_only = predict_only
        self.prediction_type = prediction_type
        self.predicted_quantity = predicted_quantity
        self.y = y
        self.nominal_capacity = nominal_capacity

    @classmethod
    def from_processed_cycler_run_file(
        cls,
        path,
        features_label="full_model",
        predict_only=False,
        predicted_quantity="cycle",
        prediction_type="multi",
        diagnostic_features=False,
    ):
        """
        Args:
            path (str): string corresponding to file path with ProcessedCyclerRun object.
            features_label (str): name of method for featurization.
            predict_only (bool): True/False to specify predict/train mode.
            predicted_quantity (str): 'cycle' or 'capacity'.
            prediction_type (str): Type of regression - 'single' vs 'multi'.
            diagnostic_features (bool): whether to compute diagnostic features.
        """
        processed_cycler_run = auto_load_processed(path)

        if features_label == "full_model":
            return cls.init_full_model(
                processed_cycler_run,
                predict_only=predict_only,
                predicted_quantity=predicted_quantity,
                diagnostic_features=diagnostic_features,
                prediction_type=prediction_type,
            )
        else:
            raise NotImplementedError

    @classmethod
    def init_full_model(
        cls,
        processed_cycler_run,
        init_pred_cycle=10,
        mid_pred_cycle=91,
        final_pred_cycle=100,
        predict_only=False,
        prediction_type="multi",
        predicted_quantity="cycle",
        diagnostic_features=False,
    ):
        """
        Generate features listed in early prediction manuscript
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): information about cycler run
            init_pred_cycle (int): index of initial cycle index used for predictions
            mid_pred_cycle (int): index of intermediate cycle index used for predictions
            final_pred_cycle (int): index of highest cycle index used for predictions
            predict_only (bool): whether or not to include cycler life in the object
            prediction_type (str): 'single': cycle life to reach 80% capacity.
                                   'multi': remaining capacity at fixed cycles
            predicted_quantity (str): quantity being predicted - cycles/capacity
            diagnostic_features (bool): whether or not to compute diagnostic features
        Returns:
            beep.featurize.DegradationPredictor: DegradationPredictor corresponding to the ProcessedCyclerRun file.
        """
        assert mid_pred_cycle > 10, "Insufficient cycles for analysis"
        assert (
            final_pred_cycle > mid_pred_cycle
        ), "Must have final_pred_cycle > mid_pred_cycle"
        i_final = final_pred_cycle - 1  # python indexing
        i_mid = mid_pred_cycle - 1
        summary = processed_cycler_run.structured_summary
        assert (
            len(processed_cycler_run.structured_summary) > final_pred_cycle
        ), "cycle count must exceed final_pred_cycle"
        cycles_to_average_over = (
            40  # For nominal capacity, use median discharge capacity of first n cycles
        )

        # Features in "nature energy" set only use discharge portion of the cycle
        if "step_type" in processed_cycler_run.structured_data.columns:
            interpolated_df = processed_cycler_run.structured_data[
                processed_cycler_run.structured_data.step_type == "discharge"
            ]
        else:
            interpolated_df = processed_cycler_run.structured_data

        X = pd.DataFrame(np.zeros((1, 20)))
        labels = []
        # Discharge capacity, cycle 2 = Q(n=2)
        X[0] = summary.discharge_capacity[1]
        labels.append("discharge_capacity_cycle_2")

        # Max discharge capacity - discharge capacity, cycle 2 = max_n(Q(n)) - Q(n=2)
        X[1] = max(
            summary.discharge_capacity[np.arange(final_pred_cycle)]
            - summary.discharge_capacity[1]
        )
        labels.append("max_discharge_capacity_difference")

        # Discharge capacity, cycle 100 = Q(n=100)
        X[2] = summary.discharge_capacity[i_final]
        labels.append("discharge_capacity_cycle_100")

        # Feature representing time-temperature integral over cycles 2 to 100
        X[3] = np.nansum(
            summary.time_temperature_integrated[np.arange(final_pred_cycle)]
        )
        labels.append("integrated_time_temperature_cycles_1:100")

        # Mean of charge times of first 5 cycles
        X[4] = np.nanmean(summary.charge_duration[1:6])
        labels.append("charge_time_cycles_1:5")

        # Descriptors based on capacity loss between cycles 10 and 100.
        Qd_final = interpolated_df.discharge_capacity[
            interpolated_df.cycle_index == i_final
        ]
        Qd_10 = interpolated_df.discharge_capacity[interpolated_df.cycle_index == 9]

        Qd_diff = Qd_final.values - Qd_10.values

        X[5] = np.log10(np.abs(np.min(Qd_diff)))  # Minimum
        labels.append("abs_min_discharge_capacity_difference_cycles_2:100")

        X[6] = np.log10(np.abs(np.mean(Qd_diff)))  # Mean
        labels.append("abs_mean_discharge_capacity_difference_cycles_2:100")

        X[7] = np.log10(np.abs(np.var(Qd_diff)))  # Variance
        labels.append("abs_variance_discharge_capacity_difference_cycles_2:100")

        X[8] = np.log10(np.abs(skew(Qd_diff)))  # Skewness
        labels.append("abs_skew_discharge_capacity_difference_cycles_2:100")

        X[9] = np.log10(np.abs(kurtosis(Qd_diff)))  # Kurtosis
        labels.append("abs_kurtosis_discharge_capacity_difference_cycles_2:100")

        X[10] = np.log10(np.abs(Qd_diff[0]))  # First difference
        labels.append("abs_first_discharge_capacity_difference_cycles_2:100")

        X[11] = max(
            summary.temperature_maximum[list(range(1, final_pred_cycle))]
        )  # Max T
        labels.append("max_temperature_cycles_1:100")

        X[12] = min(
            summary.temperature_minimum[list(range(1, final_pred_cycle))]
        )  # Min T
        labels.append("min_temperature_cycles_1:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 2 to 100

        X[13], X[14] = np.polyfit(
            list(range(1, final_pred_cycle)),
            summary.discharge_capacity[list(range(1, final_pred_cycle))],
            1,
        )

        labels.append("slope_discharge_capacity_cycle_number_2:100")
        labels.append("intercept_discharge_capacity_cycle_number_2:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 91 to 100
        X[15], X[16] = np.polyfit(
            list(range(i_mid, final_pred_cycle)),
            summary.discharge_capacity[list(range(i_mid, final_pred_cycle))],
            1,
        )
        labels.append("slope_discharge_capacity_cycle_number_91:100")
        labels.append("intercept_discharge_capacity_cycle_number_91:100")

        IR_trend = summary.dc_internal_resistance[list(range(1, final_pred_cycle))]
        if any(v == 0 for v in IR_trend):
            IR_trend[IR_trend == 0] = np.nan

        # Internal resistance minimum
        X[17] = np.nanmin(IR_trend)
        labels.append("min_internal_resistance_cycles_2:100")

        # Internal resistance at cycle 2
        X[18] = summary.dc_internal_resistance[1]
        labels.append("internal_resistance_cycle_2")

        # Internal resistance at cycle 100 - cycle 2
        X[19] = (
            summary.dc_internal_resistance[i_final] - summary.dc_internal_resistance[1]
        )
        labels.append("internal_resistance_difference_cycles_2:100")

        X.columns = labels
        if predict_only:
            y = None
        else:
            if prediction_type == "single":
                y = processed_cycler_run.get_cycle_life()
            elif prediction_type == "multi":
                if predicted_quantity == "cycle":
                    y = processed_cycler_run.capacities_to_cycles(
                        thresh_max_cap=0.98, thresh_min_cap=0.78, interval_cap=0.03
                    )
                elif predicted_quantity == "capacity":
                    y = processed_cycler_run.cycles_to_capacities()
                else:
                    raise NotImplementedError(
                        "{} predicted_quantity type not implemented".format(
                            predicted_quantity
                        )
                    )
        nominal_capacity = np.median(
            summary.discharge_capacity.iloc[0:cycles_to_average_over]
        )

        return cls(
            "full_model",
            X,
            feature_labels=labels,
            y=y,
            nominal_capacity=nominal_capacity,
            predict_only=predict_only,
            prediction_type=prediction_type,
            predicted_quantity=predicted_quantity,
        )

    def as_dict(self):
        """
        Method for dictionary serialization
        Returns:
            dict: corresponding to dictionary for serialization
        """
        obj = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "name": self.name,
            "X": self.X.to_dict("list"),
            "feature_labels": self.feature_labels,
            "predict_only": self.predict_only,
            "prediction_type": self.prediction_type,
            "nominal_capacity": self.nominal_capacity,
        }
        if isinstance(self.y, pd.DataFrame):
            obj["y"] = self.y.to_dict("list")
        else:
            obj["y"] = self.y
        return obj

    @classmethod
    def from_dict(cls, d):
        """MSONable deserialization method"""
        d["X"] = pd.DataFrame(d["X"])
        return cls(**d)


def add_file_prefix_to_path(path, prefix):
    """
    Helper function to add file prefix to path.

    Args:
        path (str): full path to file.
        prefix (str): prefix for file.

    Returns:
        str: path with prefix appended to filename.

    """
    split_path = list(os.path.split(path))
    split_path[-1] = prefix + split_path[-1]
    return os.path.join(*split_path)


def process_file_list_from_json(file_list_json, processed_dir="data-share/features/"):
    """
    Function to take a json file containing processed cycler run file locations,
    extract features, dump the processed file into a predetermined directory,
    and return a jsonable dict of feature file locations.

    Args:
        file_list_json (str): json string or json filename corresponding
            to a dictionary with a file_list attribute,
            if this string ends with ".json", a json file is assumed
            and loaded, otherwise interpreted as a json string.
        processed_dir (str): location for processed cycler run output files
            to be placed.

    Returns:
        str: json string of feature files (with key "file_list").

    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup workflow
    outputs = WorkflowOutputs()

    # Add root path to processed_dir
    processed_dir = os.path.join(
        os.environ.get("BEEP_PROCESSING_DIR", "/"), processed_dir
    )
    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)

    file_list = file_list_data["file_list"]
    run_ids = file_list_data["run_list"]
    processed_run_list = []
    processed_result_list = []
    processed_message_list = []
    processed_paths_list = []

    for path, run_id in zip(file_list, run_ids):
        logger.info("run_id=%s featurizing=%s", str(run_id), path, extra=s)
        processed_cycler_run = auto_load_processed(path)

        featurizer_classes = [
            HPPCResistanceVoltageFeatures,
            DiagnosticSummaryStats,
            DeltaQFastCharge,
            TrajectoryFastCharge,
            DiagnosticProperties,
            intracell_losses.IntracellCycles,
            intracell_losses.IntracellFeatures,
        ]

        for featurizer_class in featurizer_classes:
            if featurizer_class.class_feature_name in FEATURE_HYPERPARAMS.keys():
                params_dict = FEATURE_HYPERPARAMS[featurizer_class.class_feature_name]
            else:
                params_dict = None
            featurizer = featurizer_class.from_run(
                path, processed_dir, processed_cycler_run, params_dict
            )
            if featurizer:
                dumpfn(featurizer, featurizer.name)
                processed_paths_list.append(featurizer.name)
                processed_run_list.append(run_id)
                processed_result_list.append("success")
                processed_message_list.append({"comment": "", "error": ""})
                logger.info("Successfully generated %s", featurizer.name, extra=s)
            else:
                processed_paths_list.append(path)
                processed_run_list.append(run_id)
                processed_result_list.append("incomplete")
                processed_message_list.append(
                    {
                        "comment": "Insufficient or incorrect data for featurization",
                        "error": "",
                    }
                )
                logger.info("Unable to featurize %s", path, extra=s)

    output_data = {
        "file_list": processed_paths_list,
        "run_list": processed_run_list,
        "result_list": processed_result_list,
        "message_list": processed_message_list,
    }

    # Workflow outputs
    outputs.put_workflow_outputs_list(output_data, "featurizing")

    # Return jsonable file list
    return json.dumps(output_data)


def main():
    """
    Main function of this module, takes in arguments of an input
    and output filename corresponding to structured cycler run data
    and creates a predictor object output for analysis/ML processing

    Returns:
        None

    """
    # Parse args and construct initial cycler run
    logger.info("starting", extra=s)
    logger.info("Running version=%s", __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args["INPUT_JSON"]
        print(process_file_list_from_json(input_json), end="")
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info("finish", extra=s)

    return None


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import json
from monty.json import MSONable
from monty.serialization import loadfn
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from beep.structure.cli import auto_load_processed


class PrincipalComponents(MSONable):
    """
    PCA object.

    Attributes:
        data (pandas.DataFrame): dataframe to be decomposed using PCA.
        name (str): name for PCA instance.
        n_components (int): number of principal components to use.
        explained_variance_threshold (float): desired variance to be explained.
        pca (sklearn.pca.PCA): pca object.
    """

    def __init__(
        self,
        data,
        name="FastCharge",
        n_components=15,
        explained_variance_threshold=0.90,
    ):
        """
        Args:
            data (pandas.DataFrame): dataframe to be decomposed using PCA.
            name (str): name for PCA instance.
            n_components (int): number of principal components to use.
            explained_variance_threshold (float): desired variance to be explained.
        """
        self.data = data
        self.name = name
        self.explained_variance_threshold = explained_variance_threshold
        self.n_components = n_components
        self.scaler = StandardScaler()
        self.pca = PCA(n_components=self.n_components)
        self.fit()
        self.get_reconstruction_errors()

    @classmethod
    def from_interpolated_data(
        cls,
        file_list_json,
        name="FastCharge",
        qty_to_pca="discharge_capacity",
        pivot_column="voltage",
        cycles_to_pca=np.linspace(20, 500, 20, dtype=int),
    ):
        """
        Method to take a list of structure jsons containing interpolated capacity vs voltage,
        create a PCA object and perform fitting.

        Args:
            file_list_json (str): json string or json filename corresponding.
            name (str):
            qty_to_pca (str): string denoting quantity to pca.
            pivot_column (str): string denoting column to pivot on. For PCA of
                Q(V), pivot_column would be voltage.
            cycles_to_pca (np.ndarray): how many cycles per file to use for pca decomposition.

        Returns:
            beep.principal_components.PrincipalComponents:
        """
        return cls(
            pivot_data(file_list_json, qty_to_pca, pivot_column, cycles_to_pca), name
        )

    def as_dict(self):
        """
        Method for dictionary/json serialization.

        Returns:
            dict: object representation as dictionary.
        """
        obj = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "pca": self.pca.__dict__,
            "scaler": self.scaler.__dict__,
            "embeddings": self.embeddings,
            "reconstruction_errors": self.reconstruction_errors,
        }
        return obj

    def fit(self):
        """
        Method to scale the dataframe, run PCA and evaluate embeddings.
        """
        # Center and scale training data
        scaled_data = self.scaler.fit_transform(self.data)
        self.pca.fit(scaled_data)
        # Find minimum number of components to explain threshold amount of variance in the data.
        self.min_components = (
            np.min(
                np.where(
                    np.cumsum(self.pca.explained_variance_ratio_)
                    > self.explained_variance_threshold
                )
            )
            + 1
        )
        # Eval embeddings of training data
        self.embeddings = self.pca.transform(scaled_data)
        self.white_embeddings = (
            self.embeddings - np.mean(self.embeddings, axis=0)
        ) / np.std(self.embeddings, axis=0)
        self.reconstructions = self.scaler.inverse_transform(
            self.pca.inverse_transform(self.embeddings)
        )
        return

    def get_pca_embeddings(self, data):
        """
        Method to compute PCA embeddings on new data using the trained PCA fit.

        Args:
            data (pandas.DataFrame): data frame

        Returns:
            numpy.array: transformed to embedded space, shape (n_samples, n_components)

        """
        return self.pca.transform(self.scaler.transform(data))

    def get_pca_reconstruction(self, embeddings):
        """
        Method to inverse transform PCA embeddings to reconstruct data

        Returns:
            numpy.array, shape [n_samples, n_features]. Transformed array.
        """
        return self.scaler.inverse_transform(self.pca.inverse_transform(embeddings))

    def get_pca_decomposition_outliers(self, data, upper_quantile=95, lower_quantile=5):
        """
        Outlier detection using PCA decomposition.

        Args:
            data (pandas.DataFrame): dataframe for which outlier detection needs
                to be performed
            upper_quantile (int): upper quantile for outlier detection
            lower_quantile (int): lower quantile for outlier detection

        Returns:
            numpy.array: distances to center of PCA set
            numpy.array: boolean vector of same length as data
        """
        # Compute center of the PCA training set
        center = np.median(self.white_embeddings, axis=0)
        # Define upper and lower quantiles for detecting outliers
        q_upper, q_lower = np.percentile(
            self.white_embeddings, [upper_quantile, lower_quantile], axis=0
        )
        # Transform new data
        embeddings = self.pca.transform(self.scaler.transform(data))
        # Compute centered embeddings and distances
        white_embeddings = (embeddings - np.mean(self.embeddings, axis=0)) / np.std(
            self.embeddings, axis=0
        )
        distances = np.linalg.norm(white_embeddings - center, axis=1)
        # Flag outliers even if one of the principal components falls out of inter-quantile range
        outlier_list = (white_embeddings > q_upper).any(axis=1) | (
            white_embeddings < q_lower
        ).any(axis=1)
        return distances, outlier_list

    def get_reconstruction_errors(self):
        """
        Method to compute reconstruction errors of training dataset
        """
        self.reconstruction_errors = np.mean(
            np.abs(self.reconstructions - self.data), axis=1
        )
        return

    def get_reconstruction_error_outliers(self, data, threshold=1.5):
        """
        Get outliers based on PCA reconstruction errors.

        Args:
            data (pandas.DataFrame): dataframe for which outlier detection needs to be performed.
            threshold (float): threshold for outlier detection

        Returns:
            numpy.array: vector of same length as data.
        """
        embeddings = self.pca.transform(self.scaler.transform(data))
        reconstructions = self.scaler.inverse_transform(
            self.pca.inverse_transform(embeddings)
        )
        reconstruction_errors = np.mean(np.abs(reconstructions - data), axis=1)
        return (
            reconstruction_errors,
            reconstruction_errors > max(self.reconstruction_errors) * threshold,
        )


def pivot_data(
    file_list_json,
    qty_to_pca="discharge_capacity",
    pivot_column="voltage",
    cycles_to_pca=np.linspace(10, 100, 10, dtype=int),
):
    """
    Method to take a list of structure jsons, construct a dataframe to PCA using
    a pivoting column.

    Args:
        file_list_json (str): json string or json filename corresponding to a
            dictionary with a file_list and validity attribute, if this string
            ends with ".json", a json file is assumed.
        qty_to_pca (str): string denoting quantity to pca.
        pivot_column (str): string denoting column to pivot on. For PCA of Q(V),
            pivot_column would be voltage.
        cycles_to_pca (np.array): how many cycles per file to use for pca
            decomposition.

    Returns:
        pandas.DataFrame: pandas dataframe to PCA.
    """
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)
    file_list = file_list_data["file_list"]
    df_to_pca = pd.DataFrame()
    for file in file_list:
        processed_run = auto_load_processed(file)

        df = processed_run.structured_data[processed_run.structured_data.step_type == "discharge"]

        df = df[df.cycle_index.isin(cycles_to_pca)]
        df_to_pca = df_to_pca.append(
            df.pivot(index="cycle_index", columns=pivot_column, values=qty_to_pca),
            ignore_index=True,
        )
    return df_to_pca



"""Classes and functions for handling Maccor battery cycler data.

"""
import os
from datetime import datetime
from glob import glob

import pandas as pd
import numpy as np
import pytz

from beep import tqdm, StringIO
from beep.conversion_schemas import MACCOR_CONFIG
from beep.structure.base_eis import BEEPDatapathWithEIS, EIS


class MaccorDatapath(BEEPDatapathWithEIS):
    """Datapath for ingesting and structuring Maccor battery cycler data.

    Attributes:
        - all attributes inherited from BEEPDatapath
        - eis [MaccorEIS]: List of MaccorEIS objects, each representing an electrochemical
            impedance spectrum complete with data and metadata.
    """

    class MaccorEIS(EIS):
        """Class representing a single EIS run from Maccor data.
        """

        @classmethod
        def from_file(cls, filename):
            """Create a single MaccorEIS object from a raw EIS output file.

            Args:
                filename (str, Pathlike): file path to data.

            Returns:
                (MaccorEIS): EIS object representation of data.
            """
            with open(filename) as f:
                lines = f.readlines()
            # Parse freq sweep, method, and output filename
            freq_sweep = lines[1].split("Frequency Sweep:")[1].strip()
            freq_sweep = freq_sweep.replace("Circut", "Circuit")
            method = lines[2]
            filename = lines[3]

            # Parse start datetime and reformat in isoformat
            start = lines[6].split("Start Date:")[1].strip()
            date, time = start.split("Start Time:")
            start = ",".join([date.strip(), time.strip()])
            start = datetime.strptime(start, "%A, %B %d, %Y,%H:%M")
            start = start.isoformat()

            line_8 = lines[8].split()

            # Construct metadata dictionary
            metadata = {
                "frequency_sweep": freq_sweep,
                "method": method,
                "filename": filename,
                "start": start,
                "line_8": line_8,
            }

            data = "\n".join(lines[10:])
            data = pd.read_csv(StringIO(data), delimiter="\t")
            return cls(data=data, metadata=metadata)

    @classmethod
    def from_file(cls, path):
        """Create a MaccorDatapath file from a Maccor cycler run raw file.

        Args:
            path (str, Pathlike): file path for maccor file.

        Returns:
            (MaccorDatapath)
        """
        with open(path) as f:
            metadata_line = f.readline().strip()

        # Parse data
        data = pd.read_csv(path, delimiter="\t", skiprows=1)
        data.rename(str.lower, axis="columns", inplace=True)
        data = data.astype(MACCOR_CONFIG["data_types"])
        data.rename(MACCOR_CONFIG["data_columns"], axis="columns", inplace=True)
        data["charge_capacity"] = cls.quantity_sum(
            data, "capacity", "charge"
        )
        data["discharge_capacity"] = cls.quantity_sum(
            data, "capacity", "discharge"
        )
        data["charge_energy"] = cls.quantity_sum(data, "energy", "charge")
        data["discharge_energy"] = cls.quantity_sum(
            data, "energy", "discharge"
        )

        # Parse metadata - kinda hackish way to do it, but it works
        metadata = cls.parse_metadata(metadata_line)
        metadata = pd.DataFrame(metadata)
        _, channel_number = os.path.splitext(path)
        metadata["channel_id"] = int(channel_number.replace(".", ""))
        metadata.rename(str.lower, axis="columns", inplace=True)
        metadata.rename(MACCOR_CONFIG["metadata_fields"], axis="columns", inplace=True)
        # Note the to_dict, which scrubs numpy typing
        metadata = {col: item[0] for col, item in metadata.to_dict("list").items()}

        # standardizing time format
        data["date_time_iso"] = data["date_time"].apply(cls.correct_timestamp)

        paths = {
            "raw": path,
            "metadata": path
        }

        return cls(data, metadata, paths=paths)

    def load_eis(self, paths=None):
        """Load eis from specified paths to EIS files, or automatically detect them from
        the directory where the raw maccor data file is located.

        This method sets MaccorDatapath.eis to a list MaccorEIS objects.
        This method also updates MaccorDatapath.paths to reflect EIS paths.

        Args:
            paths((str, Pathlike) or None): Paths to Maccor EIS files. If None, will automatically
                scan the directory where the raw MaccorDatapath file was located and choose all
                files matching the Maccor EIS RegEx as EIS files.

        Returns:
            None
        """

        # Automatically find EIS from directory

        if not paths:
            # todo: replace with logging
            print(f"Looking in directory {self.paths['raw']} for EIS runs, as no paths specified to `load_eis`.")
            eis_pattern = ".*.".join(self.paths["raw"].rsplit(".", 1))
            paths = glob(eis_pattern)

        eis_runs = []
        for path in tqdm.tqdm(paths, desc="Loading EIS files..."):
            eis = self.MaccorEIS.from_file(path)
            eis_runs.append(eis)

        self.eis = eis_runs
        self.paths["eis"] = paths

        # todo: add logging for if no paths added

    @staticmethod
    def quantity_sum(data, quantity, state_type):
        """Computes non-decreasing capacity or energy (either charge or discharge)
        through multiple steps of a single cycle and resets capacity at the
        start of each new cycle. Input Maccor data resets to zero at each step.

        Args:
            data (pd.DataFrame): maccor data.
            quantity (str): capacity or energy.
            state_type (str): charge or discharge.

        Returns:
            Series: summed quantities.

        """
        state_code = MACCOR_CONFIG["{}_state_code".format(state_type)]
        quantity_agg = data['_' + quantity].where(data["_state"] == state_code, other=0, axis=0)

        # If a waveform step is present, maccor initializes waveform-specific quantities
        # that are to be used in place of '_capacity' and '_energy'

        if data['_wf_chg_cap'].notna().sum():
            if (state_type, quantity) == ('discharge', 'capacity'):
                quantity_agg = data['_wf_dis_cap'].where(data['_wf_dis_cap'].notna(), other=quantity_agg, axis=0)
            elif (state_type, quantity) == ('charge', 'capacity'):
                quantity_agg = data['_wf_chg_cap'].where(data['_wf_chg_cap'].notna(), other=quantity_agg, axis=0)
            elif (state_type, quantity) == ('discharge', 'energy'):
                quantity_agg = data['_wf_dis_e'].where(data['_wf_dis_e'].notna(), other=quantity_agg, axis=0)
            elif (state_type, quantity) == ('charge', 'energy'):
                quantity_agg = data['_wf_chg_e'].where(data['_wf_chg_e'].notna(), other=quantity_agg, axis=0)
            else:
                pass

        end_step = data["_ending_status"].apply(
            lambda x: MACCOR_CONFIG["end_step_code_min"] <= x <= MACCOR_CONFIG["end_step_code_max"]
        )
        # For waveform discharges, maccor seems to trigger ending_status within a step multiple times
        # As a fix, compute the actual step change using diff() on step_index and set end_step to be
        # a logical AND(step_change, end_step)
        is_step_change = data['step_index'].diff(periods=-1).fillna(value=0) != 0
        end_step_inds = end_step.index[np.logical_and(list(end_step), list(is_step_change))]
        # If no end steps, quantity not reset, return it without modifying
        if end_step_inds.size == 0:
            return quantity_agg

        # Initialize accumulator and beginning step slice index
        cycle_sum = 0.0
        begin_step_ind = quantity_agg.index[0] + 1
        for end_step_ind in end_step_inds:
            # Detect whether cycle changed and reset accumulator if so
            if (
                data.loc[begin_step_ind - 1, "cycle_index"]
                != data.loc[begin_step_ind, "cycle_index"]
            ):
                cycle_sum = 0.0

            # Add accumulator to current reset step
            quantity_agg[begin_step_ind:end_step_ind + 1] += cycle_sum

            # Update accumulator
            cycle_sum = quantity_agg[end_step_ind]

            # Set new step slice initial index
            begin_step_ind = end_step_ind + 1

        # Update any dangling step without an end
        last_index = quantity_agg.index[-1]
        if end_step_inds[-1] < last_index:
            quantity_agg[begin_step_ind:] += cycle_sum
        return quantity_agg

    @staticmethod
    def parse_metadata(metadata_string):
        """Parses maccor metadata string, which is annoyingly inconsistent.
        Basically just splits the string by a set of fields and creates
        a dictionary of pairs of fields and values with colons scrubbed
        from fields.

        Args:
            metadata_string (str): string corresponding to maccor metadata.

        Returns:
            dict: dictionary of metadata fields and values.

        """
        metadata_fields = [
            "Today's Date",
            "Date of Test:",
            "Filename:",
            "Procedure:",
            "Comment/Barcode:",
        ]
        metadata_values = MaccorDatapath.split_string_by_fields(metadata_string,
                                                                metadata_fields)
        metadata = {
            k.replace(":", ""): [v.strip()]
            for k, v in zip(metadata_fields, metadata_values)
        }
        return metadata

    @staticmethod
    def correct_timestamp(x):
        """Helper function with exception handling for cases where the
        maccor cycler mis-prints the datetime stamp for the row. This
        happens when data is being recorded rapidly as the date switches over
        ie. between 10/21/2019 23:59:59 and 10/22/2019 00:00:00.

        Args:
            x (str): The datetime string for maccor in format '%m/%d/%Y %H:%M:%S'

        Returns:
            datetime.Datetime: Datetime object in iso format (daylight savings aware)

        """
        pacific = pytz.timezone("US/Pacific")
        utc = pytz.timezone("UTC")
        try:
            iso = (
                pacific.localize(datetime.strptime(x, "%m/%d/%Y %H:%M:%S"),
                                 is_dst=True)
                .astimezone(utc)
                .isoformat()
            )
        except ValueError:
            x = x + " 00:00:00"
            iso = (
                pacific.localize(datetime.strptime(x, "%m/%d/%Y %H:%M:%S"),
                                 is_dst=True)
                .astimezone(utc)
                .isoformat()
            )
        return iso

    @staticmethod
    def split_string_by_fields(string, fields):
        """Helper function to split a string by a set of ordered strings,
        primarily used for Maccor metadata parsing.

        >>> MaccorDatapath.split_string_by_fields("first name: Joey  last name Montoya",
        >>>                       ["first name:", "last name"])
        ["Joey", "Montoya"]

        Args:
            string (str): string input to be split
            fields (list): list of fields to split input string by.

        Returns:
            list: substrings corresponding to the split input strings.

        """
        # A bit brittle, there's probably something more clever with recursion
        substrings = []
        init, leftovers = string.split(fields[0])
        for field in fields[1:]:
            init, leftovers = leftovers.split(field)
            substrings.append(init)
        substrings.append(leftovers)
        return substrings



import os
import hashlib
import json
from collections import OrderedDict
from .workflow import Logger, WorkflowOutputs
from .parameters_lookup import get_protocol_parameters, get_project_sequence, get_diagnostic_parameters
from .splice import MaccorSplice
from pydash import get, set_with, unset, merge


class DashOrderedDict(OrderedDict):
    """
    Nested data structure with pydash enabled
    getters and setters.  Nested values can
    be set using dot notation, e. g.

    >>> dod = DashOrderedDict()
    >>> dod.set('key1.key2', 5)
    >>> print(dod['key1']['key2'])
    >>> 5
    """

    def set(self, string, value):
        set_with(self, string, value, lambda x: OrderedDict())

    def get_path(self, string, default=None):
        return get(self, string, default=default)

    def unset(self, string):
        unset(self, string)

    def merge(self, obj):
        merge(self, obj)

    def __str__(self):
        return "{}:\n{}".format(self.__class__.__name__, json.dumps(self, indent=4))

    def __repr__(self):
        return self.__str__()


def hash_file(filename):
    """
    Utility function to hash a file

    Args:
        filename (str): name fo file to hash
    """
    with open(filename, "rb") as f:
        chunk = f.read()
    return hashlib.md5(chunk).digest()


def os_format(json_string):
    """
    Helper function to format json string into something
    that can be parsed on the command line.  For nt (windows)
    systems, uses enclosing double quotes and escaped quotes,
    for POSIX systems uses enclosing single quotes and
    no escape characters.

    Args:
        json_string (str): json string to be formatted
    """
    if os.name == "nt":
        return '"{}"'.format(json_string.replace('"', '\\"'))
    else:
        return "'{}'".format(json_string)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module and scripts for assembling a ML model training dataset using BeepFeatures objects

Options:
    -h --help        Show this screen
    --fit            <true_or_false>  [default: False] Fit model
    --version        Show version


The `dataset` script will assemble a BeepDataset object for ML model training
It stores its outputs in `/data-share/datasets/`


"""
from __future__ import division
import os
import pandas as pd
import numpy as np
from monty.json import MSONable
from monty.serialization import loadfn, dumpfn
from functools import reduce
from beep import MODULE_DIR
from beep.utils import parameters_lookup
from beep.structure.cli import auto_load_processed
from beep.featurize import (
    HPPCResistanceVoltageFeatures,
    DiagnosticProperties,
    DiagnosticSummaryStats
)
from sklearn.model_selection import train_test_split
from scipy import interpolate

FEATURE_HYPERPARAMS = loadfn(
    os.path.join(MODULE_DIR, "features/feature_hyperparameters.yaml")
)

FEATURIZER_CLASSES = [HPPCResistanceVoltageFeatures,
                      DiagnosticSummaryStats,
                      DiagnosticProperties]


class BeepDataset(MSONable):
    """
    Class corresponding to a training dataset assembled from BeepFeatures objects

    Attributes:
            name (str): name of the dataset
            data (pd.DataFrame): dataframe composed of different features concatenated column-wise and
            different runs concatenated row-wise
            metadata (list): list of metadata dicts for the different feature objects
            filenames (list): list of filenames that have atleast one of the feature objects
            feature_sets (dict): dictionary with feature class names as keys, and feature labels as values
            dataset_dir (str): path to store serialized dataset
            train_cells_parameter_dict (dict): Dictionary with keys corresponding to unique identifiers for runs that
                are part of the training dataset, and values being a dictionary of project parameters for the run
            test_cells_parameter_dict (dict): Dictionary with keys corresponding to unique identifiers for runs that
                are part of the test dataset, and values being a dictionary of project parameters for the run
            X_train (pd.DataFrame): Training dataset predictors
            X_test (pd.DataFrame): Test dataset predictors
            y_train (pd.DataFrame): Training dataset outcomes
            y_test (pd.DataFrame): Test dataset outcomes
            missing (pd.DataFrame): Feature sets that could not be found, or not be initialized because
             the ProcessedCyclerRun object did not meet the necessary validation criteria.

    """

    def __init__(self, name, data, metadata, filenames, feature_sets, dataset_dir='data-share/datasets', missing=None):

        """
        Invokes BeepDataset object

        Args:
            name (str): name of the dataset
            data (pd.DataFrame): dataframe composed of different features concatenated column-wise and
            different runs concatenated row-wise
            metadata (list): list of metadata dicts for the different feature objects
            filenames (list): list of filenames that have atleast one of the feature objects
            feature_sets (dict): dictionary with feature class names as keys, and feature labels as values
            dataset_dir (str): path to store serialized dataset
            missing (pd.DataFrame): Feature sets that could not be found, or not be initialized because
             the ProcessedCyclerRun object did not meet the necessary validation criteria.
        """
        self.name = name
        self.data = data
        self.metadata = metadata
        self.filenames = filenames
        self.feature_sets = feature_sets
        self.dataset_dir = dataset_dir
        self.train_cells_parameter_dict = {}
        self.test_cells_parameter_dict = {}
        self.X_train = None
        self.X_test = None
        self.y_train = None
        self.y_test = None
        self.missing = missing

    def as_dict(self):
        """
        Method for dictionary serialization
        Returns:
            dict: corresponding to dictionary for serialization
        """
        obj = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "name": self.name,
            "data": self.data.to_dict("list"),
            "metadata": self.metadata,
            "filenames": self.filenames,
            "dataset_dir": self.dataset_dir,
            "feature_sets": self.feature_sets,
            "missing": self.missing.to_dict("list")
        }
        return obj

    @classmethod
    def from_dict(cls, d):
        """MSONable deserialization method"""
        d["data"] = pd.DataFrame(d["data"])
        d["missing"] = pd.DataFrame(d["missing"])
        d["filenames"] = list(d["filenames"])
        return cls(**d)

    @classmethod
    def from_features(cls, name, project_list=['PreDiag'], feature_class_list=FEATURIZER_CLASSES,
                      feature_dir="data-share/features/", dataset_dir="data-share/datasets"):
        """
        Method to assemble a dataset from a list of BeepFeatures objects generated for one or more projects. 

        project_list (list): list of projects from which training data will be assembled
        feature_class_list (list): list of BeepFeatures classes
        feature_dir (str): Root directory for features. Assumes that all objects belonging to a feature class
            are stored in a folder <feature_dir>/<MyFeatureSet.class_feature_name>
        dataset_dir (str): path to store serialized dataset

        Returns:
            beep.BeepDataset object
        """
        feature_df_list = []
        metadata = []
        feature_sets = {}
        missing = pd.DataFrame(columns=['filename', 'feature_class'])

        for feature_class in feature_class_list:
            feature_path = os.path.join(feature_dir, feature_class.class_feature_name)
            if os.path.isdir(feature_path):
                feature_df = pd.DataFrame()
                for project in project_list:
                    feature_jsons = [os.path.join(feature_path, f) for f in os.listdir(feature_path) if
                                     (os.path.isfile(os.path.join(feature_path, f)) and
                                      f.startswith(project))]
                    feature_jsons.sort()
                    for feature_json in feature_jsons:
                        obj = loadfn(feature_json)
                        df = obj.X
                        df['file'] = obj.metadata['protocol'].split('.')[0]
                        # seq_num computation assumes file naming follows the convention:
                        # ProjectName_SeqNum_Channel_ObjectName.json
                        df['seq_num'] = int(os.path.basename(feature_json).split('_')[1])
                        feature_df = pd.concat([feature_df, df]).reset_index(drop=True)
                        # TODO: Need some logic for ensuring that features of a given class being concatenated
                        # row-wise have the same metadata dict

                feature_df_list.append(feature_df)
                feature_sets[feature_class.class_feature_name] = list(feature_df.columns)

            else:
                missing.loc[len(missing)] = \
                    ['', feature_class.class_feature_name]

        df = reduce(lambda x, y: pd.merge(x, y, on=['file', 'seq_num'], how='outer'), feature_df_list)
        # Outer-join used so that even partially featurized cells can be loaded into dataset
        # NaNs imputation to be done downstream by the user

        return cls(name, df, metadata, df.file.unique(), feature_sets, dataset_dir, missing)

    @classmethod
    def from_processed_cycler_runs(cls, name, project_list, processed_run_list=None,
                                   feature_class_list=FEATURIZER_CLASSES,
                                   hyperparameter_dict=None, processed_dir="data-share/structure/",
                                   feature_dir="data-share/features/",
                                   dataset_dir="data-share/datasets",
                                   parameters_path="data-share/raw/parameters"):
        """
        Method to assemble a dataset directly from a list of ProcessedCyclerRun objects

        Arguments:
            project_list (list): list of projects to featurize and combine as a training dataset
            processed_run_list: (list) list of paths to specific ProcessedCyclerRun objects to be featurized.
                If provided, this will over-ride project based looping.
            feature_class_list: list of BeepFeatures objects to invoke on the structured cycler files.
            hyperparameter_dict (dict): dictionary with keys belonging to feature_class_list, and values being
                a list of hyperparam dictionaries for that feature_class. List allows multiple instances of the
                same BeepFeatures class to be created and assembled into the training dataset
            processed_dir (str): root directory storing structure jsons (ProcessedCyclerRun objects)
            feature_dir (str): root directory for features (BeepFeatures objects)
            dataset_dir (str): location to store dataset

        Returns:
            beep.BeepDataset object
        """
        feature_sets = {}
        failed_featurizations = pd.DataFrame(columns=['filename', 'feature_class'])

        # If no metadata is provided, assume defaults
        if hyperparameter_dict is None:
            hyperparameter_dict = {}
            print('No hyperparameters specified for feature generation. Assuming defaults and proceeding.')
            for feature_class in feature_class_list:
                if feature_class.class_feature_name in FEATURE_HYPERPARAMS.keys():
                    hyperparameter_dict[feature_class.class_feature_name] = \
                        [FEATURE_HYPERPARAMS[feature_class.class_feature_name]]
                else:
                    hyperparameter_dict[feature_class.class_feature_name] = None
        else:
            for feature_class in feature_class_list:
                # if a specific feature class has a missing hyperparameter dict, initialize with defaults
                if hyperparameter_dict[feature_class.class_feature_name] is None:
                    print('Assuming default hyperparameter dictionary for',
                          feature_class.class_feature_name)
                    if feature_class.class_feature_name in FEATURE_HYPERPARAMS.keys():
                        hyperparameter_dict[feature_class.class_feature_name] = \
                            [FEATURE_HYPERPARAMS[feature_class.class_feature_name]]
                    else:
                        hyperparameter_dict[feature_class.class_feature_name] = None
                # if the provided hyperparameter dictionary has the wrong keys, raise error
                elif set(FEATURE_HYPERPARAMS[feature_class.class_feature_name].keys()) != \
                        set(hyperparameter_dict[feature_class.class_feature_name][0].keys()):
                    raise ValueError('Invalid hyperparameter dictionary for' +
                                     feature_class.class_feature_name)

        # If a list of paths to ProcessedCyclerRun objects is not provided, then use all
        # files belonging to a project in <processed_dir>.
        if processed_run_list is None:
            processed_run_list = [os.path.join(processed_dir, f)
                                  for f in os.listdir(processed_dir)
                                  for project in project_list
                                  if (os.path.isfile(os.path.join(processed_dir, f)) and
                                      f.startswith(project) and
                                      f.endswith('structure.json'))]
        # feature_df_list is a list of dataframes. Each dataframe in it
        feature_df_list = [pd.DataFrame()]*sum([len(x) for x in hyperparameter_dict.values()])

        for processed_json in processed_run_list:
            processed_cycler_run = auto_load_processed(processed_json)
            idx = 0
            for feature_class in feature_class_list:
                # For a given feature_class, loop through multiple hyperparameter combinations, if provided.
                for d in hyperparameter_dict[feature_class.class_feature_name]:
                    obj = feature_class.from_run(processed_json, feature_dir, processed_cycler_run,
                                                 d, parameters_path=parameters_path)
                    if obj:
                        df = obj.X
                        df['file'] = obj.metadata['protocol'].split('.')[0]
                        df['seq_num'] = int(os.path.basename(processed_json).split('_')[1])
                        feature_df_list[idx] = pd.concat([feature_df_list[idx], df]).reset_index(drop=True)
                    else:
                        failed_featurizations.loc[len(failed_featurizations)] = \
                            [os.path.split(processed_json)[1], feature_class.class_feature_name]
                    idx += 1

        for idx, feature_class in enumerate(feature_class_list):
            feature_sets[feature_class.class_feature_name] = list(feature_df_list[idx].columns)

        df = reduce(lambda x, y: pd.merge(x, y, on=['file', 'seq_num'], how='outer'), feature_df_list)
        return cls(name, df, hyperparameter_dict, df.file.unique(), feature_sets, dataset_dir, failed_featurizations)

    def generate_train_test_split(self, predictors=None, outcomes=None,
                                  split_by_cell=True, test_size=0.4, seed=123,
                                  parameters_path="data-share/raw/parameters"):
        """
        Method that subsets self.data into training and test datasets. Requires specification of columns to use as
        predictors and outcomes.

        Args:
            predictors (list): list of columns to use as predictors
            outcomes (list): list of columns to use as outcomes
            split_by_cell (bool): If True, train-test split on a per-run basis (self.filenames). Useful when
                there are multiple data-points per cell to avoid data-leaks between train and test data.
            seed (int): seed to ensure reproducible 'randomization'
            parameters_path (str): Root directory storing project parameter files. Assumes that parameter files
                begin with project name

        Returns:
            pd.DataFrame: X_train, X_test, y_train, y_test
        """

        if predictors is None:
            raise ValueError('Specify one or more predictor columns')

        if outcomes is None:
            raise ValueError('Specify one or more outcomes')

        np.random.seed(seed)
        if split_by_cell:
            test_cells = np.random.choice(self.filenames, int(len(self.filenames) * test_size))
            train_cells = [x for x in self.filenames if x not in test_cells]

            self.X_train = self.data.loc[self.data.file.isin(train_cells), predictors]
            self.X_test = self.data.loc[self.data.file.isin(test_cells), predictors]

            self.y_train = self.data.loc[self.data.file.isin(train_cells), outcomes]
            self.y_test = self.data.loc[self.data.file.isin(test_cells), outcomes]
        else:
            self.X_train, self.X_test, self.y_train, self.y_test = \
                train_test_split(self.data[predictors], self.data[outcomes], test_size, random_state=seed)

        if parameters_path is not None:
            self.train_cells_parameter_dict = get_parameter_dict(train_cells, parameters_path)
            self.test_cells_parameter_dict = get_parameter_dict(test_cells, parameters_path)

        return self.X_train, self.X_test, self.y_train, self.y_test

    def serialize(self):
        """
        Method to serialize dataset

        Args:
            processed_dir (dict): target directory.

        Returns:
             Path to serialized dataset

        """
        if not os.path.exists(self.dataset_dir):
            os.makedirs(self.dataset_dir)
        dumpfn(self, os.path.join(self.dataset_dir, self.name))
        return self.dataset_dir


def get_threshold_targets(dataset_diagnostic_properties,
                          cycle_type="rpt_1C",
                          metric="discharge_energy",
                          x_axis_types=None,
                          threshold=0.80,
                          filter_kinks=None,
                          extrapolate_threshold=True):
    """
    Method to generate a data frame with cycle number and throughputs at which the cell drops below a fractional metric.
    The metric being used is set by the basis variable and the value of the threshold can be varied. The cycle being
    used to evaluate the factional metric can be set be cycle_type_target. The point at which the cell crosses the
    threshold is determined with a linear interpolation between diagnostic cycles. In the event that the cell has not
    yet reached the threshold, the point at which it is expected to reach the threshold can be calculated by linear
    extrapolation from the last two diagnostic cycles. The filter kinks option allows for removal of cells with an
    abrupt change in degradation rate, possibly indicating some error or problem with the experiment.

    Args:
        dataset_diagnostic_properties (BeepDataset.data): Data attribute of dataset object created from the
            DiagnosticProperties feature
        cycle_type (str): Type of diagnostic cycle being used to measure the fractional metric
        metric (str): The metric being used for fractional capacity
        x_axis_types (list): List of column names to use for x_axis interpolation (distance to threshold)
        threshold (float): Value for the fractional metric to be considered above or below threshold
        filter_kinks (float): If set, cutoff value for the second derivative of the fractional metric (cells with an
            abrupt change in degradation rate might have something else going on). Typical value might be 0.04
        extrapolate_threshold (bool): Should threshold crossing point be extrapolated for cells that have not yet
            reached the threshold (warning: this uses a linear extrapolation from the last two diagnostic cycles)

    Returns:
         pd.DataFrame: Each row is a seq_num and contains the interpolated throughput and cycle number at which
            that particular run crossed the threshold.
            :param x_axis_types:

    """
    # Set the x axis types for the calculating the distance to the threshold
    if x_axis_types is None:
        x_axis_types = ["normalized_regular_throughput", "cycle_index"]
    threshold_values_df_list = []
    # Only use the target cycle type and basis for calculation
    cycle_type_target_df = dataset_diagnostic_properties[dataset_diagnostic_properties.cycle_type == cycle_type]
    cycle_type_target_df = cycle_type_target_df[cycle_type_target_df['metric'] == metric]

    for indx, run in enumerate(dataset_diagnostic_properties['file'].unique()):
        run_x_axis_types = x_axis_types.copy()
        # Look at one run at a time
        run_target_df = cycle_type_target_df[cycle_type_target_df['file'] == run]

        # Filter to truncate data from cells that have a sudden drop in the fractional metric
        # (something wrong with the test)
        if filter_kinks and np.any(run_target_df['fractional_metric'].diff().diff() < filter_kinks):
            last_good_cycle = run_target_df[run_target_df['fractional_metric'].diff().diff() < filter_kinks][
                'cycle_index'].min()
            run_target_df = run_target_df[run_target_df['cycle_index'] < last_good_cycle]

        x_axes = []
        for type in run_x_axis_types:
            x_axes.append(run_target_df[type])
        y_interpolation_axis = run_target_df['fractional_metric']

        # Logic around how to deal with cells that have not crossed the threshold
        if run_target_df['fractional_metric'].min() > threshold and not extrapolate_threshold:
            print(run, " has not crossed threshold and extrapolation is off")
            continue
        elif run_target_df['fractional_metric'].min() > threshold and extrapolate_threshold:
            fill_value = "extrapolate"
            bounds_error = False
            x_linspaces = []
            for x_axis in x_axes:
                y1 = y_interpolation_axis.iloc[-2]
                y2 = y_interpolation_axis.iloc[-1]
                x1 = x_axis.iloc[-2]
                x2 = x_axis.iloc[-1]
                x_thresh_extrap = (threshold - 0.1 - y1) * (x2 - x1) / (y2 - y1) + x1
                x_linspaces.append(np.linspace(x_axis.min(), x_thresh_extrap, num=1000))
        else:
            fill_value = np.nan
            bounds_error = True
            x_linspaces = []
            for x_axis in x_axes:
                x_linspaces.append(np.linspace(x_axis.min(), x_axis.max(), num=1000))

        f_axis = []
        for x_axis in x_axes:
            f_axis.append(interpolate.interp1d(x_axis, y_interpolation_axis, kind='linear',
                                               bounds_error=bounds_error, fill_value=fill_value))

        x_to_threshold = []
        for indx, x_linspace in enumerate(x_linspaces):
            crossing_array = abs(f_axis[indx](x_linspace) - threshold)
            x_to_threshold.append(x_linspace[np.argmin(crossing_array)])

        if ~(x_to_threshold[0] > 0) or ~(x_to_threshold[1] > 0):
            print(run, " does not have a positive value to threshold")
            continue

        if "normalized_regular_throughput" in run_x_axis_types:
            real_throughput_to_threshold = x_to_threshold[run_x_axis_types.index("normalized_regular_throughput")] * \
                                           run_target_df['initial_regular_throughput'].values[0]
            x_to_threshold.append(real_throughput_to_threshold)
            run_x_axis_types.append("real_regular_throughput")

        threshold_dict = {
            "file": [run],
            "seq_num": [int(run.split("_")[1])],
            'initial_regular_throughput': run_target_df['initial_regular_throughput'].values[0],
        }

        for indx, x_axis in enumerate(run_x_axis_types):
            threshold_dict[cycle_type + metric + str(threshold) + '_' + x_axis] = [x_to_threshold[indx]]

        threshold_values_df_list.append(pd.DataFrame(threshold_dict))
    threshold_targets_df = pd.concat(threshold_values_df_list)
    threshold_targets_df.reset_index(drop=True, inplace=True)
    return threshold_targets_df


def get_parameter_dict(file_list, parameters_path):
    """
    Helper function to generate a dictionary with

    Args:
        file_list (list): List of filenames from self.filenames
        parameters_path (str): Root directory storing project parameter files.

    Returns:
        Dictionary with file_list as keys, and corresponding dictionary of protocol parameters as values
    """
    d = {}  # dict allows combining two different project parameter sets into the same structure
    for file in file_list:
        param_row, _ = parameters_lookup.get_protocol_parameters(file, parameters_path)
        d[file] = param_row.to_dict('records')[0]  # to_dict('records') returns a list.
    return d



"""Classes and functions for handling Indigo battery cycler data.

"""
from datetime import datetime

import pytz
import pandas as pd

from beep.structure.base import BEEPDatapath
from beep.conversion_schemas import INDIGO_CONFIG


class IndigoDatapath(BEEPDatapath):
    """Datapath for ingesting and structuring Indigo battery cycler data.

    """

    @classmethod
    def from_file(cls, path):
        """Creates a BEEPDatapath from an raw Indigo data file.

        Args:
            path (str, Pathlike): file path to data file

        Returns:
            (IndigoDatapath)
        """

        data = pd.read_hdf(path, "time_series_data")
        metadata = dict()

        if len(list(data["cell_id"].unique())) > 1:
            raise ValueError("More than 1 cell_id exists in {}".format(path))

        metadata["indigo_cell_id"] = int(data["cell_id"].iloc[0])
        metadata["filename"] = path
        metadata["_today_datetime"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # transformations
        data = (
            data.reset_index().reset_index()
        )  # twice in case old index is stored in file
        data = data.drop(columns=["index"])
        data.rename(columns={"level_0": "data_point"}, inplace=True)
        data.loc[data.half_cycle_count % 2 == 1, "charge_capacity"] = (
                abs(data.cell_coulomb_count_c) / 3600
        )
        data.loc[data.half_cycle_count % 2 == 0, "charge_capacity"] = 0
        data.loc[data.half_cycle_count % 2 == 0, "discharge_capacity"] = (
                abs(data.cell_coulomb_count_c) / 3600
        )
        data.loc[data.half_cycle_count % 2 == 1, "discharge_capacity"] = 0
        data.loc[data.half_cycle_count % 2 == 1, "charge_energy"] = abs(
            data.cell_energy_j
        )
        data.loc[data.half_cycle_count % 2 == 0, "charge_energy"] = 0
        data.loc[data.half_cycle_count % 2 == 0, "discharge_energy"] = abs(
            data.cell_energy_j
        )
        data.loc[data.half_cycle_count % 2 == 1, "discharge_energy"] = 0
        data["internal_resistance"] = data.cell_voltage_v / data.cell_current_a
        data["date_time_iso"] = data["system_time_us"].apply(
            lambda x: datetime.utcfromtimestamp(x / 1000000)
            .replace(tzinfo=pytz.UTC)
            .isoformat()
        )

        data.rename(INDIGO_CONFIG["data_columns"], axis="columns", inplace=True)

        metadata["start_datetime"] = data.sort_values(by="system_time_us")[
            "date_time_iso"
        ].iloc[0]

        paths = {
            "raw": path,
            "metadata": path
        }

        return cls(data, metadata, paths)
